import numpy as np
from scipy.interpolate._bsplines import BSpline

from acspype import ACSDev

TEST_FILEPATH = '../dev_tools/test_files/ACS-00011_2022-10-20.dev'


def test():
    dev = ACSDev(TEST_FILEPATH)
    assert len(dev.a_wavelength) == len(dev.c_wavelength)
    assert isinstance(dev.a_offset, np.ndarray)
    assert isinstance(dev.func_a_delta_t, BSpline)



"""
This module contains helper functions for accessing and working with OOI ACS data.
The OOI has named all their ACS sensors OPTAA.
"""

from datetime import datetime
import numpy as np
import os
import pandas as pd
import re
import requests
from scipy.interpolate import make_interp_spline
import shutil
import xarray as xr


def reformat_ooi_optaa(ds: xr.Dataset) -> xr.Dataset:
    """
    Reformat an OOI OPTAA dataset to make it more compatible with acspype.

    :param ds: An OOI OPTAA dataset.
    :return: A reformatted and renamed OOI OPTAA dataset.
    """

    # Old datasets don't have qartod flags, but newer ones do. ACS data are too finicky to trust the qartod test results implicitly.
    ds = ds.drop_vars([dv for dv in ds.data_vars if 'qartod' in dv], errors = 'ignore')

    # Remove vars 99.9999% of users won't care about.
    vars_to_drop = ['driver_timestamp', 'uuid', 'provenance_uuid', 'internal_timestamp', 'ingestion_timestamp',
                    'profiler_timestamp', 'port_timestamp', 'preferred_timestamp', 'suspect_timestamp', 'raw_pressure',
                    'optical_absorption', 'beam_attenuation','checksum',
                    'meter_type','packet_type','record_length','serial_number','uuid']
    ds = ds.drop_vars(vars_to_drop, errors='ignore')

    # Rename Wavelength dimensions
    ds = ds.rename({'wavelength_a': 'a_wavelength', 'wavelength_c': 'c_wavelength'})
    a_wvl = np.unique(ds.a_wavelength)
    c_wvl = np.unique(ds.c_wavelength)
    try:
        nwvls = int(np.unique(ds.num_wavelengths))
    except:
        nwvls = int(np.unique(ds.num_wavelengths[0]))
    ds = ds.drop_vars(['a_wavelength', 'c_wavelength', 'num_wavelengths'], errors='ignore')

    # Pull out lat/lon.
    # No OOI OPTAA dataset is currently "mobile", so assign a static latitude and longitude to the entire dataset.
    lat = np.unique(ds.lat)
    if len(lat) == 1:
        lat = lat[0]
    else:
        lat = lat.tolist()
    lon = np.unique(ds.lon)
    if len(lon) == 1:
        lon = lon[0]
    else:
        lon = lon.tolist()
    ds = ds.drop_vars(['lat', 'lon'], errors='ignore')

    # Pull out deployment variable to assign as a dimension later.
    dep = np.unique(ds.deployment)
    if len(dep) > 1:
        raise ImportError('File should not have two deployments worth of data. Please manually inspect.')
    else:
        dep = int(dep.tolist()[0])
    ds = ds.drop_vars(['deployment'], errors='ignore')

    # Reformat to use time instead of obs.
    ds = ds.swap_dims({'obs': 'time'})
    ds = ds.drop_vars(['obs'], errors='ignore')
    ds = ds.drop_duplicates(dim = 'time', keep = 'first')

    # Create a new dataset with the appropriate dimensions.
    nds = xr.Dataset()
    nds = nds.assign_coords({'time': ds.time, 'a_wavelength': a_wvl, 'c_wavelength': c_wvl})
    for var in ds.data_vars:
        if var in ['a_signal_counts', 'a_reference_counts', 'optical_absorption']:
            nds[var] = (['time', 'a_wavelength'], ds[var].data)
        elif var in ['c_signal_counts', 'c_reference_counts', 'beam_attenuation']:
            nds[var] = (['time', 'c_wavelength'], ds[var].data)
        else:
            nds[var] = (['time'], ds[var].data)

    # Rename variables to match acspype recommended naming conventions.
    mapper = {'external_temp_raw': 'raw_external_temperature',
              'c_signal_counts': 'c_signal',
              'a_signal_counts': 'a_signal',
              'int_ctd_pressure': 'sea_water_pressure',
              'a_reference_counts': 'a_reference',
              'driver_timestamp': 'driver_timestamp',
              'id': 'uuid',
              'provenance': 'provenance_uuid',
              'internal_timestamp': 'internal_timestamp',
              'beam_attenuation': 'beam_attenuation',
              'profiler_timestamp': 'profiler_timestamp',
              'internal_temp_raw': 'raw_internal_temperature',
              'ingestion_timestamp': 'ingestion_timestamp',
              'c_reference_dark_counts': 'c_reference_dark',
              'port_timestamp': 'port_timestamp',
              'optical_absorption': 'optical_absorption',
              'sea_water_practical_salinity': 'sea_water_practical_salinity',
              'on_seconds': 'elapsed_time',
              'elapsed_run_time': 'elapsed_time',
              'c_signal_dark_counts': 'c_signal_dark',
              'pressure_counts': 'raw_pressure',
              'preferred_timestamp': 'preferred_timestamp',
              'a_signal_dark_counts': 'a_signal_dark',
              'a_reference_dark_counts': 'a_reference_dark',
              'c_reference_counts': 'c_reference',
              'suspect_timestamp': 'suspect_timestamp',
              'sea_water_temperature': 'sea_water_temperature'}

    for key, value in mapper.items():
        try:
            nds = nds.rename({key: value})
            nds[value].attrs = ds[key].attrs
        except:
            continue

    nds = nds.drop_vars(['uuid','provenance_uuid','raw_pressure'], errors = 'ignore')


    # Update elapsed_time if it arrives with units of seconds.
    if nds.elapsed_time.attrs['units'] == 's':
        nds['elapsed_time'] = nds.elapsed_time * 1000
        nds['elapsed_time'].attrs['units'] = 'ms'

    # Reassign lat, lon, and deployment as attributes.
    nds.attrs['latitude'] = lat
    nds.attrs['longitude'] = lon
    nds.attrs['deployment'] = dep

    # Reassign attributes
    for attr in sorted(ds.attrs):
        if ds.attrs[attr] in ['', '[]']:
            continue
        else:
            nds.attrs[attr] = ds.attrs[attr]
    nds.attrs['number_of_output_wavelengths'] = nwvls

    # Reorder variables alphabetically for convenience.
    nds = nds[sorted(nds.data_vars)]

    return nds


def _build_acpype_dev_obj(response_data, start) -> object:
    """
    Build an ACSDev-like object for processing OOI OPTAA data with _acspype.

    :param response_data: The response from an UID request to the OOI API.
    :param start: The start time of dataset, used to find the most recent calibration.
    :return: An ACSDev-like class object.
    """

    cals = response_data['calibration']
    cal_data = {}
    for cal in cals:
        if cal['name'] == 'CC_acwo':
            cname = 'a_offset'
        elif cal['name'] == 'CC_ccwo':
            cname = 'c_offset'
        elif cal['name'] == 'CC_tcarray':
            cname = 'c_delta_t'
        elif cal['name'] == 'CC_taarray':
            cname = 'a_delta_t'
        elif cal['name'] == 'CC_tcal':
            cname = 'tcal'
        elif cal['name'] == 'CC_awlngth':
            cname = 'a_wavelength'
        elif cal['name'] == 'CC_cwlngth':
            cname = 'c_wavelength'
        elif cal['name'] == 'CC_tbins':
            cname = 'tbins'
        else:
            raise ValueError(f'Unknown calibration value name: {cal["name"]}')
        subcals = cal['calData']
        matching_cals = {}
        for subcal in subcals:
            source = subcal['dataSource']
            dt_str = re.findall(r'_(\d{8})_', source)[0]
            dt = datetime.strptime(dt_str, '%Y%m%d')
            if dt <= start:
                matching_cals[dt] = subcal['value']

        cal_dates = list(matching_cals.keys())
        diffs = [start - cd for cd in cal_dates]
        matching_dt = cal_dates[np.argmin(diffs)]
        cal_data[cname] = matching_cals[matching_dt]

    class Dev:
        # Simulate an ACSDev object.

        a_wavelength = cal_data['a_wavelength']
        c_wavelength = cal_data['c_wavelength']
        tbin = cal_data['tbins']
        a_offset = cal_data['a_offset']
        c_offset = cal_data['c_offset']
        a_delta_t = cal_data['a_delta_t']
        c_delta_t = cal_data['c_delta_t']
        tcal = cal_data['tcal']
        path_length = 0.25

        # func_a_delta_t = interpolate.interp1d(cal_data['tbins'], cal_data['a_delta_t'], axis=1)
        # func_c_delta_t = interpolate.interp1d(cal_data['tbins'], cal_data['c_delta_t'], axis=1)
        func_a_delta_t = make_interp_spline(cal_data['tbins'], cal_data['a_delta_t'], k=1, axis=1)
        func_c_delta_t = make_interp_spline(cal_data['tbins'], cal_data['c_delta_t'], k=1, axis=1)

    return Dev


def get_ooi_optaa_cal(ds: xr.Dataset) -> object:
    """
    Find an OOI OPTAA calibration for a given dataset.

    :param ds: The OOI OPTAA dataset. It must contain data from a single deployment.
        It is recommended to not change the format of the file until after running this function.
    :return: A ACSDev-like object containing the calibration data.
    """

    start = pd.to_datetime(ds.time.min().values)
    asset_url = 'https://ooinet.oceanobservatories.org/api/m2m/12587/asset'
    params = {'uid': ds.attrs['AssetUniqueID']}
    response = requests.get(asset_url, params=params)
    if response.status_code == requests.codes.ok:
        if 'aintenance' in response.text:  # Check for (m)aintenance message
            raise ConnectionError('OOI API is currently down for maintenance.')
        else:
            data = response.json()
            dev = _build_acpype_dev_obj(data, start)
            return dev


def download_and_load_goldcopy(thredds_fileserver_url: str,
                               save_dir: str = 'ooi_data/') -> xr.Dataset:
    """
    Download and load a dataset from the OOI THREDDS GoldCopy catalog.
    Sometimes data accessed through the OpenDAP URL is incomplete, so downloading is recommended.

    :param thredds_fileserver_url: A THREDDS GoldCopy URL. Must contain fileServer, not dodsC.
    :param save_dir: The directory to save the dataset to.
    :return: The xarray dataset.
    """
    os.makedirs(save_dir, exist_ok=True)
    filename = os.path.basename(thredds_fileserver_url)
    fp = os.path.join(save_dir, filename)
    if os.path.isfile(fp):
        ds = xr.open_dataset(fp)
        return ds
    else:
        with requests.get(thredds_fileserver_url, stream=True) as req:
            with open(fp, 'wb') as fileobj:
                shutil.copyfileobj(req.raw, fileobj)
        if os.path.isfile(fp):
            ds = xr.open_dataset(fp)
            return ds
        else:
            raise FileNotFoundError



"""
This module contains functions for working with raw ACS byte packets and converting them into more readable data
structures. The ACSStream class has wrapper functions around the functions in this module. Users may want to use these
functions separately from the ACSStream class if reprocessing from a file that contains raw packets,
such as the WETView generated .dat files or OOI CSPP .acs files.
"""

import numpy as np
from struct import unpack_from

from .core import WVL_BYTE_OFFSET, PACKET_HEAD, PACKET_TAIL, PAD_BYTE, LPR
from .structures import ACSPacket, ParsedPacket, DeviceCalibratedPacket, TSCorrectedPacket
from .dev import ACSDev
from .tscor import ACSTSCor
from .processing import (convert_sn_hexdec, convert_sn_str, compute_internal_temperature,
                                compute_external_temperature, compute_uncorrected, compute_measured,
                                find_discontinuity_index, compute_discontinuity_offset, apply_discontinuity_offset,
                                ts_correction, zero_shift_correction)


def unpack_packet(full_packet: bytes | bytearray) -> tuple[tuple, int]:
    """
    Unpack a full ACS packet into raw bytes and the checksum. This function is intended for use on a single packet.
        If attempting to implement with Xarray, you will need to create a separate function or attempt to wrap this
        function in an xarray.apply_ufunc.
    
    :param full_packet: A bytes or bytesarray object containing the full ACS packet, including the registration
        bytes and the pad byte.
    :return: A tuple containing the raw bytes and the checksum.
    """

    [nwvls] = unpack_from('B', full_packet, offset=WVL_BYTE_OFFSET)  # The number of wavelengths is byte 31.
    remaining_packet_size = int(nwvls * 2 * 2)
    full_packet_descriptor = PACKET_HEAD + f"{remaining_packet_size}H" + PACKET_TAIL
    pad_byte_idx = full_packet.rfind(PAD_BYTE)
    checksum_bytes = full_packet[pad_byte_idx - 2:pad_byte_idx]
    checksum = int(unpack_from('!H', checksum_bytes)[0])
    raw = unpack_from(full_packet_descriptor, full_packet)  # Note: The padbyte is not returned when unpacking.
    return raw, checksum


def parse_packet(acs_packet: ACSPacket) -> ParsedPacket:
    """
    Parse an ACS packet into a ParsedPacket object. This function is intended for use on a single ACSPacket object.
        If attempting to implement with Xarray, you will need to create a separate function or attempt to wrap this
        function in an xr.apply_ufunc.
    
    :param acs_packet: An ACSPacket object containing timestamp and the full packet data. Note: Users looking to 
        only use this to parse the packet without caring about the daq_time can use an ACSPacket with a dummy daq_time.
    :return: A ParsedPacket object containing the parsed data.
    """

    raw, checksum = unpack_packet(acs_packet.full_packet)
    nwvls = raw[LPR + 13]
    parsed_packet = ParsedPacket(
        daq_time=acs_packet.daq_time,
        record_length=raw[LPR + 0],
        packet_type=raw[LPR + 1],
        # reserved_1 = raw[LPR + 2],  # reserved_1 is not used for anything at this time.
        sn_int=raw[LPR + 3],
        a_reference_dark=raw[LPR + 4],
        # raw_pressure = raw[LPR + 5], # raw_pressure is deprecated and no longer used.
        a_signal_dark=raw[LPR + 6],
        raw_external_temperature=raw[LPR + 7],
        raw_internal_temperature=raw[LPR + 8],
        c_reference_dark=raw[LPR + 9],
        c_signal_dark=raw[LPR + 10],
        elapsed_time=raw[LPR + 11],
        # reserved_2= raw[LPR + 12], # reserved_2 is not used for anything at this time.
        number_of_wavelengths=nwvls,
        c_reference=raw[LPR + 14:LPR + 14 + (nwvls * 4):4],
        a_reference=raw[LPR + 15:LPR + 15 + (nwvls * 4):4],
        c_signal=raw[LPR + 16:LPR + 16 + (nwvls * 4):4],
        a_signal=raw[LPR + 17:LPR + 17 + (nwvls * 4):4],
        checksum=checksum,
    )
    return parsed_packet


def calibrate_packet(parsed_packet: ParsedPacket, dev: ACSDev) -> DeviceCalibratedPacket:
    """
    Calibrate a parsed packet using the device calibration parameters. This function is intended for use on a single
        ParsedPacket object. If attempting to implement with Xarray, you will need to create a separate function or
        attempt to wrap this function in an xarray.apply_ufunc.
        
    :param parsed_packet: A ParsedPacket object containing the parsed data. 
    :param dev: An ACSDev-like object containing the device calibration parameters.
    :return: A DeviceCalibratedPacket object containing data corrected/calibrated to the values in the ACSDev object.
    """
    internal_temperature = compute_internal_temperature(parsed_packet.raw_internal_temperature)
    external_temperature = compute_external_temperature(parsed_packet.raw_external_temperature)

    a_uncorrected = compute_uncorrected(parsed_packet.a_signal, parsed_packet.a_reference, dev.path_length)
    c_uncorrected = compute_uncorrected(parsed_packet.c_signal, parsed_packet.c_reference, dev.path_length)

    disc_idx = find_discontinuity_index(dev.a_wavelength, dev.c_wavelength)
    a_m_discontinuity = compute_measured(a_uncorrected, internal_temperature, dev.a_offset, dev.func_a_delta_t)
    c_m_discontinuity = compute_measured(c_uncorrected, internal_temperature, dev.c_offset, dev.func_c_delta_t)

    a_disc_offset = compute_discontinuity_offset(a_m_discontinuity, dev.a_wavelength, disc_idx, 'a_wavelength')
    c_disc_offset = compute_discontinuity_offset(c_m_discontinuity, dev.c_wavelength, disc_idx, 'c_wavelength')

    a_m = apply_discontinuity_offset(a_m_discontinuity, a_disc_offset, disc_idx, 'a_wavelength')
    c_m = apply_discontinuity_offset(c_m_discontinuity, c_disc_offset, disc_idx, 'c_wavelength')

    dev_cal_packet = DeviceCalibratedPacket(
        daq_time=parsed_packet.daq_time,
        a_wavelength=dev.a_wavelength,
        c_wavelength=dev.c_wavelength,
        sn_hexdec=convert_sn_hexdec(parsed_packet.sn_int),
        serial_number=convert_sn_str(parsed_packet.sn_int),
        internal_temperature=internal_temperature,
        external_temperature=external_temperature,
        a_uncorrected=a_uncorrected,
        c_uncorrected=c_uncorrected,
        discontinuity_index=disc_idx,
        a_discontinuity_offset=a_disc_offset,
        c_discontinuity_offset=c_disc_offset,
        a_m_discontinuity=a_m_discontinuity,
        c_m_discontinuity=c_m_discontinuity,
        a_m=a_m,
        c_m=c_m,
    )
    return dev_cal_packet


def ts_correct_packet(device_calibrated_packet: DeviceCalibratedPacket,
                      temperature: float,
                      salinity: float,
                      dev: ACSDev) -> TSCorrectedPacket:
    """
    Apply TS-correction to a device calibrated packet.

    :param device_calibrated_packet: A DeviceCalibratedPacket object containing the device calibrated data.
    :param temperature: The temperature or conservative temperature of the water sample in degrees Celsius.
    :param salinity: The practical or absolute salinity of the water sample.
    :param dev: A ACSDev-like object containing the device calibration parameters.
    :return: A TSCorrectedPacket object containing data that has been corrected
        for dependence on temperature and salinity.
    """

    # Import the TSCor Coefficients. Absorption and attenuation use different coefficients.
    tscor = ACSTSCor().to_xarray()
    psi_s_a = tscor.psi_s_a.sel(wavelength=np.array(device_calibrated_packet.a_wavelength), method='nearest')
    a_psi_t = tscor.psi_t.sel(wavelength=np.array(device_calibrated_packet.a_wavelength), method='nearest')

    psi_s_c = tscor.psi_s_c.sel(wavelength=np.array(device_calibrated_packet.c_wavelength), method='nearest')
    c_psi_t = tscor.psi_t.sel(wavelength=np.array(device_calibrated_packet.c_wavelength), method='nearest')

    a_mts = ts_correction(device_calibrated_packet.a_m, temperature, salinity, a_psi_t.values, psi_s_a.values, dev.tcal)
    c_mts = ts_correction(device_calibrated_packet.c_m, temperature, salinity, c_psi_t.values, psi_s_c.values, dev.tcal)

    a_mts = zero_shift_correction(a_mts)  # Zero shift correction is applied to the a_mts values.
    c_mts = zero_shift_correction(c_mts)

    ts_cor_packet = TSCorrectedPacket(
        daq_time=device_calibrated_packet.daq_time,
        a_wavelength=device_calibrated_packet.a_wavelength,
        c_wavelength=device_calibrated_packet.c_wavelength,
        temperature=temperature,
        salinity=salinity,
        a_mts=a_mts,
        c_mts=c_mts
    )
    return ts_cor_packet



"""
This module contains functions that are beneficial to use alongside acspype, but do not support acspype directly.
"""

import time
import serial
import serial.tools.list_ports

from acspype.core import PACKET_REGISTRATION

def list_available_ports() -> list:
    """
    List available serial ports for the native operating system.

    :return: A list of available serial port names.
    """

    available_ports = [v.name for v in serial.tools.list_ports.comports()]
    return available_ports


def find_acs_port(baudrate: int = 115200,
                  timeout: int = 1,
                  check_length: int = 1) -> str:
    """
    Iterate through available serial ports and check the incoming data for the ACS registration bytes.
    The first port with ACS registration bytes is returned. This function has not been tested with multiple ACS sensors
    connected to the same computer.

    :param baudrate: The baudrate for the ACS connection. Default is 115200 bps and does not need to be changed.
    :param timeout: The timeout for the serial connection. Default is 1 second.
    :param check_length: The amount of time in seconds to collect data from each serial port in
        hopes of receiving the ACS registration bytes.
    :return: The operating system serial port as a string. Can be used by pyserial for future connections.
    """

    available_ports = list_available_ports()
    for port in available_ports:
        try:
            with serial.Serial(port=port, baudrate=baudrate, timeout=timeout) as ser:
                time.sleep(check_length)
                incoming = ser.read(ser.in_waiting)
            if PACKET_REGISTRATION in incoming:
                return port
        except:  # Bare exceptions don't follow PEP 8: E772, but do we really care for this use case?
            continue
    raise ConnectionAbortedError('No ACS detected. Is the sensor connected and on?')




import numpy as np
import xarray as xr

from acspype import ACSDev, ACSTSCor
import acspype.processing as acsproc

TEST_FILEPATH = '../dev_tools/test_files/TEST_SHIMADA_202405.nc'
TEST_DEV = '../dev_tools/test_files/ACS-00412_2023-05-10.dev'


def test():
    dev = ACSDev(TEST_DEV)
    acs = xr.open_dataset(TEST_FILEPATH)
    tscor = ACSTSCor().to_xarray()
    assert np.all(dev.a_wavelength == acs.a_wavelength)

    # Compute Temperatures from ACS Thermistors
    acs['internal_temperature'] = acsproc.compute_internal_temperature(acs.raw_internal_temperature)

    # Compute Uncorrected Absorption and Attenuation
    acs['a_uncorrected'] = acsproc.compute_uncorrected(acs.a_signal, acs.a_reference, dev.path_length)
    acs['c_uncorrected'] = acsproc.compute_uncorrected(acs.c_signal, acs.c_reference, dev.path_length)

    # Compute Measured Absorption and Attenuation
    acs['a_m'] = acsproc.compute_measured(acs.a_uncorrected, acs.internal_temperature, dev.a_offset, dev.func_a_delta_t)
    acs['c_m'] = acsproc.compute_measured(acs.c_uncorrected, acs.internal_temperature, dev.c_offset, dev.func_c_delta_t)

    # Discontinuity Correction
    discontinuity_index = acsproc.find_discontinuity_index(acs.a_wavelength, acs.c_wavelength)
    acs['a_m'], acs['a_discontinuity_offset'] = acsproc.discontinuity_correction(acs.a_m,
                                                                                 discontinuity_index,
                                                                                 'a_wavelength')
    acs['c_m'], acs['c_discontinuity_offset'] = acsproc.discontinuity_correction(acs.c_m,
                                                                                 discontinuity_index,
                                                                                 'c_wavelength')

    # Temperature and Salinity Correction
    psi_s_a = tscor.psi_s_a.sel(wavelength=np.array(acs.a_wavelength), method='nearest')
    psi_t = tscor.psi_t.sel(wavelength=np.array(acs.a_wavelength), method='nearest')
    acs['a_mts'] = acsproc.ts_correction(acs.a_m, acs.sea_water_temperature, acs.sea_water_practical_salinity, psi_t,
                                         psi_s_a, dev.tcal)

    psi_s_c = tscor.psi_s_c.sel(wavelength=np.array(acs.c_wavelength), method='nearest')
    psi_t = tscor.psi_t.sel(wavelength=np.array(acs.c_wavelength), method='nearest')
    acs['c_mts'] = acsproc.ts_correction(acs.c_m, acs.sea_water_temperature, acs.sea_water_practical_salinity, psi_t,
                                         psi_s_c, dev.tcal)

    # Zero Shift Correction
    acs['a_mts'] = acsproc.zero_shift_correction(acs.a_mts)
    acs['c_mts'] = acsproc.zero_shift_correction(acs.c_mts)

    acs = acsproc.interpolate_common_wavelengths(acs, 'a_wavelength', 'c_wavelength', 'wavelength')

    ref_a = acs.a_mts.sel(wavelength=715)
    acs['a_mts_baseline'] = acsproc.baseline_scattering_correction(acs.a_mts, ref_a)

    acs['a_mts_fixed'] = acsproc.fixed_scattering_correction(acs.a_mts, acs.c_mts)

    ref_a = acs.a_mts.sel(wavelength=715)
    ref_c = acs.c_mts.sel(wavelength=715)
    acs['a_mts_proportional'] = acsproc.proportional_scattering_correction(acs.a_mts, acs.c_mts, ref_a, ref_c)

    gelbstoff = acs.where(acs.seawater_state == 1, drop=True)

    assert gelbstoff.seawater_state.max() == 1



"""This module contains functions for performing checks on xarray.Datasets that contain ACS data."""

import numpy as np
from numpy.typing import ArrayLike


def _check_dimensions(dimensions: list) -> UserWarning | None:
    """
    Check the dimensions of the xr.Dataset to ensure that 'time' is included.
    acspype relies on the 'time' dimensions to be present in the xr.Dataset.

    :param dimensions: A list of dimension for the xr.Dataset.
    :return: No errors if the dimension checks pass, otherwise raises a ValueError.
    """

    if "time" not in dimensions:
        UserWarning("It is strongly recommended to use 'time' as the time dimension for an ACS dataset.")
    if "a_wavelength" not in dimensions and "wavelength" not in dimensions:
        UserWarning("It is strongly recommended to use 'a_wavelength' as the absorption wavelength dimension for an "
                    "ACS dataset.")
    elif "c_wavelength" not in dimensions and "wavelength" not in dimensions:
        UserWarning("It is strongly recommended to use 'c_wavelength' as the attenuation wavelength dimension for an "
                    "ACS dataset.")
    elif "wavelength" not in dimensions and "a_wavelength" not in dimensions and "c_wavelength" not in dimensions:
        UserWarning("It is strongly recommended to use 'wavelength' as the the common wavelength dimension for "
                    "the absorption and attenuation coefficients for an ACS dataset.")
    else:
        return None


def _compare_wavelengths(wavelength_bins_1: ArrayLike[float],
                         wavelength_bins_2: ArrayLike[float]) -> ValueError | None:
    """
    Compare two sets of wavelengths to ensure they are the same.

    :param wavelength_bins_1: An array of wavelengths that represent wavelengths from an ACS.
    :param wavelength_bins_2: An array of wavelengths that represent wavelengths from an ACS.
    :return: No errors if the wavelengths are the same, otherwise raises a ValueError.
    """

    if np.all(wavelength_bins_1 != wavelength_bins_2):
        raise ValueError("Mismatch between wavelength sets.")
    else:
        return None



import numpy as np
import os
import xarray as xr

from .ts4cor import TS4COR
from .ts4cor_sigma import TS4COR_SIGMA


class ACSTSCor:
    """A wrapper class for the TS4COR and TS4COR_SIGMA dictionaries that are included with acspype."""

    def __init__(self):
        self.wavelengths = tuple(TS4COR.keys())

        self.psi_s_a = list(map(lambda x: x[1]['psi_s_a'], TS4COR.items()))
        self.psi_s_c = list(map(lambda x: x[1]['psi_s_c'], TS4COR.items()))
        self.psi_t = list(map(lambda x: x[1]['psi_t'], TS4COR.items()))

        self.sigma_psi_s_a= list(map(lambda x:x[1]['sigma_psi_s_a'], TS4COR_SIGMA.items()))
        self.sigma_psi_s_c= list(map(lambda x:x[1]['sigma_psi_s_c'], TS4COR_SIGMA.items()))
        self.sigma_psi_t= list(map(lambda x:x[1]['sigma_psi_t'], TS4COR_SIGMA.items()))

        self.method = 'Sullivan et al., 2006'

    def to_xarray(self):
        """
        Convert the TS4COR class attributes to an xarray.Dataset.

        :return: A dimensioned xarray.Dataset.
        """
        ds = xr.Dataset()
        ds = ds.assign_coords({'wavelength': np.array(self.wavelengths)})
        ds['psi_t'] = (['wavelength'], np.array(self.psi_t))
        ds['psi_s_c'] = (['wavelength'], np.array(self.psi_s_c))
        ds['psi_s_a'] = (['wavelength'], np.array(self.psi_s_a))
        ds['sigma_psi_t'] = (['wavelength'], np.array(self.sigma_psi_t))
        ds['sigma_psi_s_c'] = (['wavelength'], np.array(self.sigma_psi_s_c))
        ds['sigma_psi_s_a'] = (['wavelength'], np.array(self.sigma_psi_s_a))
        ds.attrs['tscor_data'] = self.method
        return ds


class ACSTS4CorReader:
    """
    A class for parsing ACS TS4.cor files and putting them into a format that is easier to work with for larger or
    multiple file datasets.

    This class can be used if users want to import the TS4.cor file themselves and don't want to use values
    interpolated by acspype.
    """

    def __init__(self, filepath: str) -> None:
        """
        Parse the .cor file and assign data as attributes.

        :param filepath: The filepath of the TS4.cor file.
        """

        self.filepath = os.path.normpath(filepath)
        self.__read_cor()
        self.__parse_lines()
        self.method = 'Sullivan et al., 2006'

    def __read_cor(self) -> None:
        """
        Read .cor file and store lines as a class attribute.

        :return: None
        """

        with open(self.filepath, 'r') as _file:
            self._lines = _file.readlines()

    def __parse_lines(self) -> None:
        """
        Parse the lines of the .cor file to get correction information.

        :return: None
        """

        wavelengths = []
        psi_t = []
        psi_s_c = []
        psi_s_a = []
        for line in self._lines:
            line_data = line.split('\t')
            line_data = [v.replace('\n', '') for v in line_data]
            line_data = [v.replace(' ', '') for v in line_data]
            if line_data == ['']:
                break
            line_data = [float(v) for v in line_data]
            wavelengths.append(line_data[0])
            psi_t.append(line_data[1])
            psi_s_c.append(line_data[2])
            psi_s_a.append(line_data[3])
        if len(wavelengths) != len(psi_t) != len(psi_s_c) != len(psi_s_a):
            raise ValueError('Mismatch in length of TS4cor file.')
        else:
            self.wavelengths = np.array(wavelengths)
            self.psi_t = np.array(psi_t)
            self.psi_s_c = np.array(psi_s_c)
            self.psi_s_a = np.array(psi_s_a)

    def to_xarray(self):
        """
        Convert the TS4COR class attributes to an xarray.Dataset.

        :return: A dimensioned xarray.Dataset.
        """
        ds = xr.Dataset()
        ds = ds.assign_coords({'wavelength': np.array(self.wavelengths)})
        ds['psi_t'] = (['wavelength'], np.array(self.psi_t))
        ds['psi_s_c'] = (['wavelength'], np.array(self.psi_s_c))
        ds['psi_s_a'] = (['wavelength'], np.array(self.psi_s_a))

        ds.attrs['tscor_data'] = self.method
        return ds



"""
This module contains functions and classes for parsing and accessing data from ACS .dev files.
Currently, only structure version 3 is supported.
"""

from datetime import datetime
import numpy as np
import re
import xarray as xr
from scipy.interpolate import make_interp_spline

from .core import NUM_PAT


class ACSDev:
    """
    A class for parsing ACS .dev files and putting them into a format that is easier to work with for larger or
    multiple file datasets.

    Generally, users will not call individual functions, but rather use the class to obtain attributes, which are
    created at class instantiation or convert the data to an xarray dataset using the to_xarray function.
    """

    def __init__(self, filepath: str) -> None:
        """
        Run the following functions at instantiation to parse the .dev file and store the info as class attributes.

        :param filepath: The filepath to the .dev file.
        :return: None
        """

        self._filepath = filepath
        self.__read_dev()
        self.__parse_metadata()
        self.__parse_tbins()
        self.__parse_offsets()
        self.__build_interp_funcs()
        self.__check_parse()

    def __read_dev(self) -> None:
        """
        Import the .dev file as a text file.
        The file contents are stored as a list of strings in the class attribute self._content.

        :return: None
        """

        with open(self._filepath, 'r') as _file:
            self._content = _file.readlines()

    def __parse_metadata(self) -> None:
        """
        Parse the .dev file for individual sensor metadata.
        Sensor specific metadata are stored as class attributes.

        :return: None
        """

        metadata_lines = [line for line in self._content if 'C and A offset' not in line]
        for line in metadata_lines:
            if 'ACS Meter' in line:
                self.sensor_type = re.findall('(.*?)\n', line)[0]
            elif 'Serial' in line:
                self.sn_hexdec = re.findall('(.*?)\t', line)[0]
                self.serial_number = 'ACS-' + str(int(self.sn_hexdec[-6:], 16)).zfill(5)
            elif 'structure version' in line:
                self.structure_version = int(re.findall(f'({NUM_PAT})\t', line)[0])
            elif 'tcal' in line or 'Tcal' in line:
                self.tcal, self.ical = [float(v) for v in re.findall(f': ({NUM_PAT}) C', line)]
                cal_date_str = re.findall('file on (.*?)[.]', line)[0].replace(' ', '')
                try:  # Sometimes the file date is entered as yyyy or yy. This should handle both cases.
                    self.cal_date = datetime.strptime(cal_date_str, '%m/%d/%Y').strftime('%Y-%m-%d')
                except:
                    self.cal_date = datetime.strptime(cal_date_str, '%m/%d/%y').strftime('%Y-%m-%d')
            elif 'Depth calibration' in line:
                (self.depth_offset,
                 self.depth_scale_factor) = [float(v) for v in re.findall(f'({NUM_PAT})', line)]
            elif 'Baud' in line:
                self.baudrate = int(re.findall(f'({NUM_PAT})\t', line)[0])
            elif 'Path' in line:
                self.path_length = float(re.findall(f'({NUM_PAT})\t', line)[0])
            elif 'wavelengths' in line:
                self.num_wavelength = int(re.findall(f'({NUM_PAT})\t', line)[0])
            elif 'number of temperature bins' in line:
                self.num_tbin = int(re.findall(f'({NUM_PAT})\t', line)[0])
            elif 'maxANoise' in line:
                (self.max_a_noise, self.max_c_noise, self.max_a_nonconform, self.max_c_nonconform,
                 self.max_a_difference, self.max_c_difference, self.min_a_counts,
                 self.min_c_counts, self.min_r_counts, self.max_temp_sd,
                 self.max_depth_sd) = [float(v) for v in re.findall(f'({NUM_PAT})\t', line)]

    def __parse_tbins(self) -> None:
        """
        Parse the .dev file for temperature bin information.

        :return: None
        """
        tbin_line = [line for line in self._content if '; temperature bins' in line][0]
        tbins = tbin_line.split('\t')
        tbins = [v for v in tbins if v]  # Toss empty strings.
        tbins = [v for v in tbins if v != '\n']  # Toss newline characters.
        self.tbin = np.array([float(v) for v in tbins if 'temperature bins' not in v])  # Convert to float.

    def __parse_offsets(self) -> None:
        """
        Parse the .dev file for a and c offsets. Data are saved as class attributes for access at a later time.

        :return: None
        """

        offset_lines = [line for line in self._content if 'C and A offset' in line]

        # Create holder arrays to loop over and append data to.
        c_wvls = []
        a_wvls = []
        c_offs = []
        a_offs = []
        c_deltas = []
        a_deltas = []
        wavelength_color_schemes = []

        for line in offset_lines:
            offsets, c_delta, a_delta = line.split('\t\t')[:-1]
            c_wvl, a_wvl, wvl_color, c_off, a_off = offsets.split('\t')

            # Convert strings to proper pythonic datatypes.
            c_wvl = float(c_wvl.replace('C', ''))
            a_wvl = float(a_wvl.replace('A', ''))
            c_off = float(c_off)
            a_off = float(a_off)
            c_delta = [float(v) for v in c_delta.split('\t')]
            a_delta = [float(v) for v in a_delta.split('\t')]

            # Append files to holder arrays.
            c_wvls.append(c_wvl)
            a_wvls.append(a_wvl)
            c_offs.append(c_off)
            a_offs.append(a_off)
            c_deltas.append(c_delta)
            a_deltas.append(a_delta)
            wavelength_color_schemes.append(wvl_color)

        # Convert holder arrays to numpy arrays.
        self.c_wavelength = np.array(c_wvls)
        self.a_wavelength = np.array(a_wvls)
        self.c_offset = np.array(c_offs)
        self.a_offset = np.array(a_offs)
        self.c_delta_t = np.array(c_deltas)
        self.a_delta_t = np.array(a_deltas)
        self.wavelength_color_schemes = wavelength_color_schemes

    def __build_interp_funcs(self) -> None:
        """
        Build interpolation functions for the a and c delta_t values and store as class attributes.

        Note: In SciPy v 1.15.0, scipy.interpolate.interp1d has become legacy and it is recommended that other code
        be used. 1st degree make_interp_spline has been implemented as a replacement. Technically, legacy code is not
        considered deprecated, so users could overwrite the function class attributes with interp1d if preferred.

        :return: None
        """
        # self.func_a_delta_t = interpolate.interp1d(self.tbin, self.a_delta_t, axis=1)
        # self.func_c_delta_t = interpolate.interp1d(self.tbin, self.c_delta_t, axis=1)

        self.func_a_delta_t = make_interp_spline(self.tbin, self.a_delta_t, k=1, axis=1)
        self.func_c_delta_t = make_interp_spline(self.tbin, self.c_delta_t, k=1, axis=1)
        self.delta_t_interp_method = 'scipy.interpolate.make_interp_spline'

    def __check_parse(self) -> None:
        """
        Verify that the shape of the data is as expected.

        :return: None
        """

        if len(self.a_wavelength) != self.num_wavelength:
            raise ValueError('Mismatch between number of wavelengths extracted for A and expected from file.'
                             'Please verify the .dev file integrity.')
        if len(self.c_wavelength) != self.num_wavelength:
            raise ValueError('Mismatch between number of wavelengths extracted for C and expected from file.'
                             'Please verify the .dev file integrity.')
        if len(self.c_wavelength) != len(self.a_wavelength):
            raise ValueError('Mismatch between number of wavelengths extracted for A and C.'
                             'Please verify the .dev file integrity.')
        if np.array(self.a_delta_t).shape != (len(self.a_wavelength), self.num_tbin):
            raise ValueError('Mismatch between length of A wavelengths and number of temperature bins.'
                             'Please verify the .dev file integrity.')
        if np.array(self.c_delta_t).shape != (len(self.a_wavelength), self.num_tbin):
            raise ValueError('Mismatch between length of C wavelengths and number of temperature bins.'
                             'Please verify the .dev file integrity.')

    def to_xarray(self) -> xr.Dataset:
        """
        Convert the parsed .dev file files to an xarray dataset

        :return: An appropriately dimensioned xarray dataset containing device file files.
        """
        ds = xr.Dataset()
        ds = ds.assign_coords({'a_wavelength': self.a_wavelength,
                               'c_wavelength': self.c_wavelength,
                               'temperature_bin': self.tbin})

        ds['a_offset'] = (['a_wavelength'], self.a_offset)
        ds['a_delta_t'] = (['a_wavelength', 'temperature_bin'], self.a_delta_t)

        ds['c_offset'] = (['c_wavelength'], np.array(self.c_offset))
        ds['c_delta_t'] = (['c_wavelength', 'temperature_bin'], self.c_delta_t)

        ds.attrs['device_filepath'] = self._filepath
        ds.attrs['sensor_type'] = self.sensor_type
        ds.attrs['serial_number_hexdec'] = self.sn_hexdec
        ds.attrs['serial_number'] = self.serial_number
        ds.attrs['device_file_structure_version'] = self.structure_version
        ds.attrs['tcal'] = self.tcal
        ds.attrs['ical'] = self.ical
        ds.attrs['calibration_date'] = self.cal_date
        ds.attrs['depth_offset'] = self.depth_offset
        ds.attrs['depth_scale_factor'] = self.depth_scale_factor
        ds.attrs['baudrate'] = self.baudrate
        ds.attrs['path_length'] = self.path_length
        ds.attrs['number_of_wavelength_bins'] = self.num_wavelength
        ds.attrs['number_of_temperature_bins'] = self.num_tbin
        ds.attrs['max_a_noise'] = self.max_a_noise
        ds.attrs['max_c_noise'] = self.max_c_noise
        ds.attrs['max_a_nonconform'] = self.max_a_nonconform
        ds.attrs['max_c_nonconform'] = self.max_c_nonconform
        ds.attrs['max_a_difference'] = self.max_a_difference
        ds.attrs['max_c_difference'] = self.max_c_difference
        ds.attrs['min_a_counts'] = self.min_a_counts
        ds.attrs['min_c_counts'] = self.min_c_counts
        ds.attrs['min_r_counts'] = self.min_r_counts
        ds.attrs['max_temp_sd'] = self.max_temp_sd
        ds.attrs['max_depth_sd'] = self.max_depth_sd
        return ds



from .dev import ACSDev
from .tscor import ACSTSCor
from .stream import ACSStream
from .structures import (ACSPacket,
                         ParsedPacket,
                         DeviceCalibratedPacket,
                         TSCorrectedPacket)


"""
This module contains functions that are used to process data from the ACS in real-time or as a batch process.

Most functions are intended to be used with scalar values/1D arrays (e.g. data produced in real-time) or
xarray DataArrays (e.g. mass processing of an ACS deployment), or both.
Scalar/1D inputs are intended to be used in the processing of a single packet or spectrum from the ACS.
xr.DataArray implementation is intended to be used in the processing of multiple spectra already in an xarray.Dataset.
For computations that are not immediately builtin to xarray,the xarray.apply_ufunc function is used to
apply the function that is denoted with a prepended underscore.

It is strongly recommended that your xarray.Dataset contain the coordinates/dimensions of
    time, a_wavelength, c_wavelength

For a list of recommended coordinate/variable names for ACS data products, please see
https://github.com/IanTBlack/acspype/blob/main/info/NAMING_CONVENTIONS.md
"""

from collections.abc import Callable
import numpy as np
from numpy.typing import NDArray
from scipy.interpolate import CubicSpline
from uncertainties import unumpy as unp
import xarray as xr

from .core import SPECS

def _convert_sn_hexdec(sn_int: int) -> str:
    """
    Converts an integer representing an ACS serial number to the hexadecimal representation.
    Application to an array of serial numbers requires this function to be vectorized. See convert_sn_hexdec.

    :param sn_int: An integer representing the ACS serial number.
    :return: The hexadecimal representation of the serial number, not really as readable as the integer representation.
    """

    sn_hexdec = hex(sn_int)
    return sn_hexdec


def convert_sn_hexdec(sn_int: int | xr.DataArray) -> str | xr.DataArray:
    """
    Converts an integer representing an ACS serial number to the hexadecimal representation.
    The function vectorizes _convert_sn_hexdec for xr.DataArray inputs.

    :param sn_int: An integer or xr.DataArray of integers representing the ACS serial number.
    :return: The hexadecimal representation of the serial number, not really as readable as the integer representation.
    """

    if not isinstance(sn_int, xr.DataArray):
        sn_hexdec = _convert_sn_hexdec(sn_int)
    else:
        sn_hexdec = xr.apply_ufunc(_convert_sn_hexdec, sn_int,
                                   input_core_dims=[[]],
                                   output_core_dims=[[]],
                                   vectorize=True)

        # Assign attributes if output is a xr.DataArray.
        sn_hexdec.attrs['ancillary_variables'] = sn_int.name
    return sn_hexdec


def _convert_sn_str(sn_int: int) -> str:
    """
    Converts an integer representing an ACS serial number to a string representation.
    Application to an array of serial numbers requires this function to be vectorized. See convert_sn_str.

    :param sn_int: An integer representing the ACS serial number.
    :return: The string representation of the serial number, zero-padded for consistent string length into the future.
    :note: The string serial number should match that of the serial number sticker on the ACS.
    """

    sn_hexdec = convert_sn_hexdec(sn_int)
    sn_str = 'ACS-' + str(int(sn_hexdec[-6:], 16)).zfill(5)  # sn string becomes ACS-XXXXX. Zero padded to 5 digits.
    return sn_str


def convert_sn_str(sn_int: int | xr.DataArray) -> str | xr.DataArray:
    """
    Converts an integer representing an ACS serial number to a string representation.
    This function vectorizes _convert_sn_str for xr.DataArray inputs.

    :param sn_int: An integer or xr.DataArray of integers representing the ACS serial number.
    :return: The string representation of the serial number.
    """

    if not isinstance(sn_int, xr.DataArray):
        return _convert_sn_str(sn_int)
    else:
        sn_str = xr.apply_ufunc(_convert_sn_str, sn_int,
                                input_core_dims=[[]],
                                output_core_dims=[[]],
                                vectorize=True)

        # Assign attributes if output is an xr.DataArray.
        sn_str.attrs['ancillary_variables'] = sn_int.name
        return sn_str

def compute_internal_temperature(counts: int | xr.DataArray,
                                 return_uncertainty: bool = False,
                                 unc: float = SPECS.INTERNAL_TEMPERATURE.UNC) -> float | xr.DataArray:
    """
    Compute the internal temperature of the ACS from the counts of the internal temperature sensor.

    :param counts: Values of the internal temperature sensor, in counts.
    :param return_uncertainty: If True, return the uncertainty of the internal temperature sensor as an uncertainties.unumpy object.
    :param unc: The unc of the internal temperature sensor.
    :return: A float value if the input was a scalar counts value or an xr.DataArray of the internal temperature
        if input as an xr.DataArray. Units of return are degrees Celsius.
    """

    a = 0.00093135
    b = 0.000221631
    c = 0.000000125741
    d = 273.15
    volts = 5 * counts / 65535
    resistance = 10000 * volts / (4.516 - volts)
    internal_temperature = 1 / (a + b * np.log(resistance) + c * np.log(resistance) ** 3) - d

    if return_uncertainty is False:
        if not isinstance(counts, xr.DataArray):
            internal_temperature = float(internal_temperature)

        # Assign attributes if output is an xr.DataArray.
        if isinstance(internal_temperature, xr.DataArray):
            internal_temperature.attrs['ancillary_variables'] = counts.name
        return internal_temperature
    else:
        internal_temperature = unp.uarray(internal_temperature, unc)
    return internal_temperature


def compute_external_temperature(counts: int | xr.DataArray,
                                 return_uncertainty: bool = False,
                                 unc: float = SPECS.INTERNAL_TEMPERATURE.UNC) -> float | xr.DataArray:
    """
    Compute the external temperature of the ACS from the counts of the external temperature sensor.
    Not used in processing if the sensor is not placed in-situ or if there is a better _source that is indicative
    of the temperature of the water sampled by the ACS.

    :param counts: The counts of the external temperature sensor.
    :return: The external temperature in degrees Celsius.
    """

    a = -7.1023317e-13
    b = 7.09341920e-08
    c = -3.87065673e-03
    d = 95.8241397
    external_temperature = a * counts ** 3 + b * counts ** 2 + c * counts + d
    if return_uncertainty is False:
        if not isinstance(counts, xr.DataArray):
            external_temperature = float(external_temperature)

        # Assign attributes if output is an xr.DataArray.
        if isinstance(external_temperature, xr.DataArray):
            external_temperature.attrs['ancillary_variables'] = counts.name
    else:
        external_temperature = unp.uarray(external_temperature, unc)
    return external_temperature


def compute_depth(counts: int | xr.DataArray,
                  scale_factor: float,
                  offset: float,
                  return_uncertainty: bool = False,
                  unc: float = SPECS.PRESSURE.UNC) -> float | xr.DataArray:
    """
    Compute depth for ACS instruments that have a pressure sensor.
    The depth is calculated from a scale factor and offset found in the .dev file.
    Formula: depth = scale_factor * counts + offset
        The offset and scale_factor are in the Depth calibration line in the dev file, in
        that particular order.

    :param counts: The pressure counts from the ACS.
    :param scale_factor: The scale factor found in the ACS dev file.
    :param offset: The offset found in the ACS dev file.
    :return: The depth in meters.
    """
    pressure = scale_factor * counts + offset

    if return_uncertainty is False:
        if not isinstance(counts, xr.DataArray):
            pressure = float(pressure)

        if isinstance(pressure, xr.DataArray):
            pressure.attrs['ancillary_variables'] = counts.name
            pressure.attrs['depth_scale_factor'] = scale_factor
            pressure.attrs['depth_offset'] = offset
    else:
        pressure = unp.uarray(pressure, unc)
    return pressure

def compute_uncorrected(signal_counts: tuple[int, ...] | NDArray[float] | xr.DataArray,
                        reference_counts: tuple[int, ...] | NDArray[float] | xr.DataArray,
                        path_length: float = 0.25,
                        return_uncertainty: bool = False,
                        unc: float = SPECS.ACS.UNC_LONG) -> NDArray[float] | xr.DataArray:
    """
    Compute the uncorrected coefficient from the signal and reference counts.

    :param signal_counts: The absorption or attenuation channel signal counts.
    :param reference_counts: The absorption or attenuation channel reference counts.
    :param path_length: The path length of the ACS in meters. May be denoted by "x" in equations and literature.
        Default is 0.25m, but it is recommended to input the path length that is in the device file.
    :return: Uncorrected absorption or attenuation coefficient in m^-1.
    """

    if not isinstance(signal_counts, xr.DataArray):
        signal_counts = np.array(signal_counts)
        reference_counts = np.array(reference_counts)

    uncorr = (1 / path_length) * np.log(signal_counts / reference_counts)

    if return_uncertainty is False:
        # Assign attributes if output is an xr.DataArray.
        if isinstance(uncorr, xr.DataArray):
            uncorr.attrs['ancillary_variables'] = [signal_counts.name, reference_counts.name]
            uncorr.attrs['path_length'] = path_length
    else:
        uncorr = unp.uarray(uncorr, unc)
    return uncorr


def compute_measured(uncorrected: NDArray[float] | xr.DataArray,
                     internal_temperature: float | xr.DataArray,
                     offset: NDArray[float],
                     func_delta_t: Callable,
                     return_uncertainty: bool = False) -> NDArray[float] | xr.DataArray:
    """
    Compute the measured absorption or attenuation coefficient from the uncorrected coefficient
    and the internal temperature.

    :param uncorrected: The uncorrected absorption or attenuation coefficient.
        Typically computed via compute_uncorrected().
    :param internal_temperature: The internal temperature of the ACS in degrees Celsius.
        Typically computed via compute_internal_temperature().
    :param offset: The coefficient offset from the device file.
    :param func_delta_t: The interpolation function for correcting for internal temperature variation.
        This function is built by default when using an ACSDev object.
    :return: Measured absorption or attenuation coefficient in m^-1, corrected for internal temperature variation.
    """

    if return_uncertainty is False:
        if not isinstance(uncorrected, xr.DataArray):
            measured = (offset - uncorrected) - func_delta_t(internal_temperature)
        else:
            # Something weird happens when using BSpline, so the transposition is necessary to get the right shape.
            measured = (offset - uncorrected) - func_delta_t(internal_temperature).T

        if isinstance(measured, xr.DataArray):
            # Assign attributes if output is an xr.DataArray.

            measured.attrs['ancillary_variables'] = [uncorrected.name, internal_temperature.name]
            measured.attrs['func_delta_t'] = 'scipy.interpolate.make_interp_spline'
            measured.attrs['func_delta_t_k_value'] = '1'

    else:

        int_temp_vals = unp.nominal_values(internal_temperature) # Split the uncertainties object
        #int_temp_sd = unp.std_devs(internal_temperature)
        interp_int_temp = func_delta_t(int_temp_vals) #Interpolate the temperature.

        measured = (offset - uncorrected) - interp_int_temp.T

    return measured


def find_discontinuity_index(a_wavelength: NDArray[float],
                             c_wavelength: NDArray[float],
                             min_wvl: float = 550.0,
                             max_wvl: float = 600.0) -> int:
    """
    Find the wavelength index of the discontinuity in the absorption and attenuation spectra. The discontinuity is
    in the range of 535-600nm.

    :param a_wavelength: The absorption wavelengths of the ACS.
    :param c_wavelength: The attenuation wavelengths of the ACS.
    :param min_wvl: Default is 535 nm. Can be adjusted if the discontinuity is not in the expected range.
    :param max_wvl: Default is 600 nm. Can be adjusted if the discontinuity is not in the expected range.
    :return: The index of the discontinuity in the absorption and attenuation spectra.
    """

    a_wvl = np.where((a_wavelength < min_wvl) | (a_wavelength > max_wvl), np.nan, a_wavelength)
    c_wvl = np.where((c_wavelength < min_wvl) | (c_wavelength > max_wvl), np.nan, c_wavelength)
    discontinuity_index = int(np.nanargmin(np.diff(a_wvl) + np.diff(c_wvl)))
    return discontinuity_index


def _compute_discontinuity_offset(measured: NDArray[float],
                                  wavelength: NDArray[float],
                                  disc_idx: int) -> float:
    """
    Compute the discontinuity offset for a measured coefficient.
    compute_discontinuity_offset provides vectorization for this function.

    :param measured: Absorption or attenuation coefficient.
    :param wavelength: The wavelengths of the absorption or attenuation spectra.
    :param disc_idx: The index of the discontinuity in the absorption and attenuation spectra.
    :return: The offset of the discontinuity in the absorption and attenuation spectra as an integer.
    """
    x = wavelength[disc_idx - 2:disc_idx + 1]
    y = measured[disc_idx - 2:disc_idx + 1]
    if np.any(np.isinf(y)) or np.any(np.isnan(y)):
        return np.nan
    else:
        cubic_spline = CubicSpline(x, y)
        interp = cubic_spline(wavelength[disc_idx + 1], extrapolate=True)
        offset = float(np.round((interp - measured[disc_idx + 1]), 5))
        return offset


def compute_discontinuity_offset(measured: NDArray[float] | xr.DataArray,
                                 wavelength: NDArray[float] | xr.DataArray,
                                 disc_idx: int,
                                 wavelength_dim: str) -> float | xr.DataArray:
    """
    Compute the scalar discontinuity offset for a measured coefficient.

    :param measured: Absorption or attenuation coefficient.
    :param wavelength: The wavelengths of the absorption or attenuation spectra.
    :param disc_idx: The index of the discontinuity in the absorption and attenuation spectra.
    :param wavelength_dim: The name of the wavelength dimension if inputting an xarray DataArray.
    :return: The discontinuity offset of the absorption and attenuation spectra as a float.
    """

    if not isinstance(measured, xr.DataArray):
        try:
            disc_off = _compute_discontinuity_offset(measured, wavelength, disc_idx)
        except:

            disc_off = np.apply_along_axis(_compute_discontinuity_offset, axis = 1,arr = unp.nominal_values(measured), wavelength = wavelength, disc_idx = disc_idx)

    elif isinstance(measured, xr.DataArray):
        disc_off = xr.apply_ufunc(_compute_discontinuity_offset, measured,
                                  kwargs={'wavelength': measured[wavelength_dim].values,
                                          'disc_idx': disc_idx},
                                  input_core_dims=[[wavelength_dim]],
                                  output_core_dims=[[]],
                                  vectorize=True)

        # Assign attributes if output is an xr.DataArray.
        disc_off.attrs['ancillary_variables'] = measured.name
        disc_off.attrs['wavelength_dimension'] = wavelength_dim
        disc_off.attrs['discontinuity_index'] = disc_idx
    return disc_off


def _apply_discontinuity_offset(measured: NDArray[float],
                                disc_off: float,
                                disc_idx: int,
                                shift_method: str = 'halfway') -> NDArray[float]:
    """
    Apply a discontinuity offset to a measured coefficient.
    Note: This function is not vectorized. The input measured array is copied to prevent memory conflicts.

    :param measured: Measured absorption or attenuation coefficient.
    :param disc_off: The discontinuity offset of the absorption and attenuation spectra.
    :param disc_idx: The index of the discontinuity in the absorption and attenuation spectra.
    :return: The spectrum with the offset applied.
    """

    _measured = np.copy(measured)
    if shift_method == 'first':
        _measured[:disc_idx + 1] = _measured[:disc_idx + 1] - disc_off  # First half
    elif shift_method == 'second':
        _measured[disc_idx + 1:] = _measured[disc_idx + 1:] + disc_off # Second half.
    elif shift_method == 'halfway':
        half_off = disc_off / 2
        _measured[:disc_idx + 1] = _measured[:disc_idx + 1] - half_off  # First half
        _measured[disc_idx + 1:] = _measured[disc_idx + 1:] + half_off  # Second half
    return _measured


def apply_discontinuity_offset(measured: NDArray[float] | xr.DataArray,
                               disc_off: float | NDArray[float],
                               disc_idx: int, wavelength_dim: str,
                               shift_method: str = 'halfway') -> NDArray[float] | xr.DataArray:
    """
    Apply a discontinuity offset to a measured coefficient. Vectorized version of _apply_discontinuity_offset, where
    applicable.

    :param measured: Measured absorption or attenuation coefficient.
    :param disc_off: The discontinuity offset of the absorption and attenuation spectra.
    :param disc_idx: The index of the discontinuity in the absorption and attenuation spectra.
    :param wavelength_dim: The name of the wavelength dimension if inputting an xarray DataArray.
    :return: The spectrum with the offset applied.
    """
    if not isinstance(measured, xr.DataArray):
        try:
            disc_applied = _apply_discontinuity_offset(measured, disc_off, disc_idx)
        except:
            vals = unp.nominal_values(measured)
            std_devs = unp.std_devs(measured)

            corr_vals = np.array([_apply_discontinuity_offset(vals[i], disc_off[i], disc_idx, shift_method) for i in range(len(disc_off))])

            # TODO: Account for uncertainty in application of discontinuity correction.


            disc_applied = unp.uarray(corr_vals, std_devs)

    else:
        disc_applied = xr.apply_ufunc(_apply_discontinuity_offset, measured, disc_off,
                                      kwargs={'disc_idx': disc_idx,
                                              'shift_method': shift_method},
                                      input_core_dims=[[wavelength_dim], []],
                                      output_core_dims=[[wavelength_dim]],
                                      vectorize=True)

        # Assign attributes if output is an xr.DataArray.
        disc_applied.attrs['ancillary_variables'] = measured.name
        disc_applied.attrs['wavelength_dimension'] = wavelength_dim
        disc_applied.attrs['discontinuity_index'] = disc_idx
        disc_applied.attrs['discontinuity_offset_applied'] = 'True'
    return disc_applied


def discontinuity_correction(measured: xr.DataArray,
                             discontinuity_index: int,
                             wavelength_dim: str,
                             shift_method: str = 'halfway') -> tuple[xr.DataArray, xr.DataArray]:
    """
    Compute the discontinuity offset and apply it to the measured coefficient.
    NOTE: This function only works with xarray DataArrays.

    :param measured: The measured absorption or attenuation coefficient.
    :param discontinuity_index: The index of the discontinuity in the absorption and attenuation spectra.
    :param wavelength_dim: The wavelength dimension of the measured coefficient if inputting an xarray DataArray.
    :return: A tuple of xr.DataArrays (corrected measured coefficient, offset)
    """

    disc_offset = compute_discontinuity_offset(measured, measured[wavelength_dim].values,
                                               discontinuity_index, wavelength_dim)
    disc_applied = apply_discontinuity_offset(measured, disc_offset,
                                              discontinuity_index,
                                              wavelength_dim,
                                              shift_method)

    disc_applied.attrs['discontinuity_corrected'] = 'True'
    disc_applied.attrs['discontinuity_shift_method'] = shift_method
    return disc_applied, disc_offset


def ts_correction(measured: NDArray[float] | xr.DataArray,
                  temperature: float | xr.DataArray,
                  salinity: float | xr.DataArray,
                  psi_temperature: NDArray[float] | xr.DataArray,
                  psi_salinity: NDArray[float] | xr.DataArray,
                  tcal: float,
                  sigma_psi_temperature: NDArray[float] | xr.DataArray | None = None,
                  sigma_psi_salinity: NDArray[float] | xr.DataArray | None = None,
                  return_uncertainty: bool = False,
                  unc_temperature: float = 0.0,
                  unc_salinity: float = 0.0) -> NDArray[float] | xr.DataArray:
    """
    Correct the measured absorption or attenuation coefficient for temperature and salinity.
    This function works on singletons and xarray DataArrays.

    :param measured: The measured absorption or attenuation coefficient.
    :param temperature: The temperature of the water in degrees Celsius.
    :param salinity: The salinity of the water in PSU.
    :param psi_temperature: The temperature coefficients from ACSTSCor,
        subset by wavelength equivalent to the corresponding measured coefficient wavelengths.
    :param psi_salinity: The temperature coefficients from ACSTSCor,
        subset by wavelength equivalent to the corresponding measured coefficient wavelengths.
    :param tcal: The tcal value from the device file.
    :return: TS-corrected absorption or attenuation coefficient in m^-1.
    """
    psi_t = psi_temperature
    psi_s = psi_salinity

    if return_uncertainty is False:
        if not isinstance(measured, xr.DataArray):
            dt = temperature - tcal
            s = salinity
            mts = measured - ((psi_t * dt) + (psi_s * s))
        else:
            dt, psi_t = np.meshgrid(temperature - tcal, psi_temperature)
            s, psi_s = np.meshgrid(salinity, psi_salinity)
            mts = measured - ((psi_t.T * dt.T) + (psi_s.T * s.T))

        if isinstance(mts, xr.DataArray):
            # Assign attributes if output is an xr.DataArray.
            mts.attrs['ancillary_variables'] = [measured.name, temperature.name, salinity.name, psi_temperature.name,
                                                psi_salinity.name]
            mts.attrs['tcal'] = tcal
    else:
        dt = unp.uarray(temperature, unc_temperature) - tcal
        s = unp.uarray(salinity, unc_salinity)

        if sigma_psi_temperature is not None and sigma_psi_salinity is not None:
            psi_t = unp.uarray(psi_t, sigma_psi_temperature.values)
            psi_s = unp.uarray(psi_s, sigma_psi_salinity.values)
        else:
            raise ValueError('For uncertainty propagation, the standard deviations for the wavelengths of interest from Sullivan et al. 2006 need to be provided.')

        dt, psi_t = np.meshgrid(dt, psi_t)
        s, psi_s = np.meshgrid(s, psi_s)

        mts = measured - ((psi_t.T * dt.T) + (psi_s.T * s.T))

    return mts


def zero_shift_correction(mts: NDArray[float] | xr.DataArray) -> NDArray[float] | xr.DataArray:
    """
    Shift values between [-0.005, 0] to 0. This function works both for 1D arrays and xarray DataArrays.

    :param mts: The TS-corrected absorption or attenuation coefficient
    :return: The shifted TS-corrected absorption or attenuation coefficient.
    """
    # try:
    #     vals = unp.nominal_values(mts)
    #     std_devs = unp.std_devs(mts)
    # except:
    #     vals = mts
    #     std_devs = None

    vals = mts
    if not isinstance(vals, xr.DataArray):
        mts = np.where((vals >= -0.005) & (vals < 0), 0, vals)
        # if std_devs is not None:
        #     mts = unp.uarray(mts, std_devs)
    else:
        mts = vals.where((vals > 0) | (vals <= -0.005), 0)
        # Assign attributes if output is an xr.DataArray.
        mts.attrs['zero_shifted'] = "True"
    return mts


def interpolate_common_wavelengths(ds: xr.Dataset,
                                   a_wavelength_dim: str = 'a_wavelength',
                                   c_wavelength_dim: str = 'c_wavelength',
                                   new_wavelength_dim: str = 'wavelength',
                                   wavelength_range: list | str = 'infer',
                                   step: int = 1, ) -> xr.Dataset:
    """
    This function interpolates the absorption and attenuation spectra to a common wavelength range and step size.
    Only works on xarray Datasets.
    Applies interpolation to all variables on the a_wavelength and c_wavelength dimensions.

    :param ds: The input dataset with two wavelength dimensions, one for absorption and the other for attenuation.
    :param a_wavelength_dim: The name of the absorption wavelength dimension.
    :param c_wavelength_dim: The name of the attenuation wavelength dimension.
    :param new_wavelength_dim: The name of the new wavelength dimension to be created.
        Default is 'wavelength', which is the recommended name for the common wavelength dimension.
    :param wavelength_range: The wavelength range to interpolate along.
        Default is 'infer', which will use the minimums and maximums of the absorption and attenuation wavelengths to
        identify a common range for both channels.
    :param step: The wavelength step size to interpolate to. Default is 1 nm.
    :return: A dataset interpolated to the common wavelength range and step size. If there are variables that you do not
        want to interpolate along (e.g. a_signal, a_reference, c_signal, c_reference), it is recommended that you subset
        before interpolation.
    """

    if wavelength_range == 'infer':
        min_wvl = np.ceil(max(ds[a_wavelength_dim].min(), ds[c_wavelength_dim].min()))  # Max of a and c wvl mins.
        max_wvl = np.floor(min(ds[a_wavelength_dim].max(), ds[c_wavelength_dim].max()))  # Min of a and c wvl maxs.
    else:
        min_wvl, max_wvl = wavelength_range
    wvls = np.arange(min_wvl, max_wvl, step)
    cds = ds.interp({a_wavelength_dim: wvls, c_wavelength_dim: wvls})  # Interpolate to step size.
    cds = cds.reset_index([a_wavelength_dim,
                           c_wavelength_dim], drop=True).assign_coords(wavelength=wvls).rename(
        {a_wavelength_dim: new_wavelength_dim,
         c_wavelength_dim: new_wavelength_dim})

    cds.attrs['wavelength_range'] = [float(min_wvl), float(max_wvl)]
    cds.attrs['interpolation_step'] = step
    return cds


def baseline_scattering_correction(a_mts: NDArray[float] | xr.DataArray,
                                   reference_a: float | xr.DataArray) -> NDArray[float] | xr.DataArray:
    """
    Perform baseline scattering correction on the absorption coefficient.
    This is the baseline method from Zaneveld et al. (1994).

    :param a_mts: The TS-corrected absorption coefficient across all wavelengths.
    :param reference_a: The TS-corrected absorption coefficient at a reference wavelength.
    :return: The baseline scattering-corrected absorption coefficient.
    """

    scatcorr = a_mts - reference_a

    if isinstance(scatcorr, xr.DataArray):  # Assign attributes if output is an xarray.DataArray.
        scatcorr.attrs['ancillary_variables'] = a_mts.name
        scatcorr.attrs['scattering_correction_method'] = 'Baseline, Zaneveld et al. (1994)'
        scatcorr.attrs['scattering_correction_method_doi'] = 'https://doi.org/10.1117/12.190095'
    return scatcorr


def fixed_scattering_correction(a_mts: NDArray[float] | xr.DataArray,
                                c_mts: NDArray[float] | xr.DataArray,
                                epsilon: float = 0.14) -> NDArray[float] | xr.DataArray:
    """
    Perform fixed scattering correction on the absorption coefficient.
    This is the fixed correction method from Pegau et al. 2003.

    :param a_mts: TS-corrected absorption.
    :param c_mts: TS-corrected attenuation.
    :param epsilon: A default value of 0.14 is used for biological particle dominant waters.
        A value of 0.18 is recommended for waters dominated by sediments.
    :return: Fixed scattering corrected absorption coefficient.
    """

    scatcorr = a_mts - epsilon * (c_mts - a_mts)

    if isinstance(scatcorr, xr.DataArray):  # Assign attributes if output is an xarray.DataArray.
        scatcorr.attrs['ancillary_variables'] = [a_mts.name, c_mts.name]
        scatcorr.attrs['epsilon'] = epsilon
        scatcorr.attrs['scattering_correction_method'] = 'Fixed, Pegau et al. (2003)'
        scatcorr.attrs['scattering_correction_method_doi'] = ('https://ntrs.nasa.gov/api/citations/'
                                                              '20030093642/downloads/20030093642.pdf')
    return scatcorr


def proportional_scattering_correction(a_mts: NDArray[float] | xr.DataArray,
                                       c_mts: NDArray[float] | xr.DataArray,
                                       reference_a: NDArray[float] | xr.DataArray,
                                       reference_c: NDArray[float] | xr.DataArray) -> NDArray[float] | xr.DataArray:
    """
    Perform proportional scattering correction on the absorption coefficient.
    This is the proportional method from Zaneveld et al. (1994).

    :param a_mts: TS-corrected absorption.
    :param c_mts: TS-corrected attenuation.
    :param reference_a: The TS-corrected absorption coefficient at a reference wavelength. Historically 715nm.
    :param reference_c: The TS-corrected attenuation coefficient at a reference wavelength.  Historically 715nm.
    :return: Proportional scattering corrected absorption coefficient.
    """

    scatcorr = a_mts - ((reference_a / (reference_c - reference_a)) * (c_mts - a_mts))

    if isinstance(scatcorr, xr.DataArray):  # Assign attributes if output is an xarray.DataArray.
        scatcorr.attrs['ancillary_variables'] = [a_mts.name, c_mts.name]
        scatcorr.attrs['scattering_correction_method'] = 'Proportional, Zaneveld et al. (1994)'
        scatcorr.attrs['scattering_correction_method_doi'] = 'https://doi.org/10.1117/12.190095'
    return scatcorr


def baseline_plus_scattering_correction(a_mts: xr.DataArray,
                                        a_mts_715: xr.DataArray) -> xr.DataArray:
    """
    Perform baseline plus scattering correction for the absorption coefficient.
    This method originates from Rottgers et al. (2013). https://doi.org/10.1016/j.mio.2013.11.001
    True absorption at 715nm is empirically determined in Equation 5.

    :param a_mts: The TS-corrected absorption coefficient.
    :param a_mts_715: The absorption at the reference wavelength. Historically, 715nm is used.
    :return: Baseline plus scattering corrected absorption coefficient.
    """

    true_a_715 = 0.212 * np.power(a_mts_715, 1.135)  # Equation 5 from Rottgers et al. (2013)
    scatcorr = a_mts - (a_mts_715 - true_a_715)  # Equation 3 from Rottgers et al. (2013)

    if isinstance(scatcorr, xr.DataArray):  # Assign attributes if output is an xarray.DataArray.
        scatcorr.attrs['ancillary_variables'] = a_mts.name
        scatcorr.attrs['scattering_correction_method'] = 'Baseline+, Rottgers et al. (2013)'
        scatcorr.attrs['scattering_correction_method_doi'] = 'https://doi.org/10.1016/j.mio.2013.11.001'
    return scatcorr


def proportional_plus_scattering_correction(a_mts: xr.DataArray,
                                            c_mts: xr.DataArray,
                                            a_mts_715: xr.DataArray,
                                            c_mts_715: xr.DataArray,
                                            e_c: float = 0.56) -> xr.DataArray:
    """
    Perform proportional plus scattering correction for the absorption coefficient.
    This method originates from Rottgers et al. (2013). https://doi.org/10.1016/j.mio.2013.11.001
    True absorption at 715nm is empirically determined in Equation 5.  The value e_c represents attenuation correction.

    This function only works with xarray DataArrays that share a common 'wavelength' dimension, thus the dataset
    must be linearly interpolated to a common wavelength range and bin size.

    :param a_mts: The TS-corrected absorption coefficient.
    :param c_mts: The TS-corrected attenuation coefficient.
    :param a_mts_715: The absorption at the reference wavelength. Historically, 715nm is used.
    :param c_mts_715: The attenuation at the reference wavelength. Historically, 715nm is used.
    :param e_c: The attenuation correction factor. Normally it is 1 (no attenuation correction),
        but Rottgers et al. (2013) suggest a value of 0.56 when running wavelength independent correction (which this
        function does).
    :return: Proportional plus scattering corrected absorption coefficient.
    """

    true_a_715 = 0.212 * np.power(a_mts_715, 1.135)  # EQ5 from Rottgers et al. (2013)
    scatcorr = a_mts - (a_mts_715 - true_a_715) * (
            ((1 / e_c) * c_mts - a_mts) / ((1 / e_c) * c_mts_715 - true_a_715))  # EQ4

    if isinstance(scatcorr, xr.DataArray):  # Assign attributes if output is an xarray.DataArray.
        scatcorr.attrs['ancillary_variables'] = a_mts.name, c_mts.name
        scatcorr.attrs['reference_wavelength'] = 715
        scatcorr.attrs['scattering_correction_method'] = 'Proportional+, Rottgers et al. (2013)'
        scatcorr.attrs['scattering_correction_method_doi'] = 'https://doi.org/10.1016/j.mio.2013.11.001'
    return scatcorr


# ----------- Experimental and Untested Functions ----------- #

def _estimate_reference_wavelength_index(a_spectra: np.array) -> int | float:
    """
    Estimate the index of the reference wavelength of an absorption spectra. This function finds the first
    wavelength in the last 1/5 wavelength bins that is greater than and nearest to zero.
    If the entire spectra is NaN, then a wvl_idx of -999 is returned.

    :param a_spectra: A 1D array of absorption spectra.
    :return: The index of the reference wavelength.
    """
    num_wvls = len(a_spectra)
    if np.isnan(num_wvls):
        return -999
    elif np.all(np.isnan(a_spectra)):
        return -999
    len_non_reds = len(a_spectra[:int(num_wvls * 4 / 5)])
    reds = a_spectra[int(num_wvls * 4 / 5):]
    reds_pos = np.where(reds <= 0, np.nan, reds)
    if np.all(np.isnan(reds_pos)):
        return -999
    else:
        reds_wvl_idx = np.nanargmin(reds_pos)
        wvl_idx = int(len_non_reds + reds_wvl_idx)
        return wvl_idx


def estimate_reference_wavelength(a_mts: xr.DataArray, wavelength_dim: str) -> float:
    """
    Estimate a reference wavelength for an ND absorption dataset.
    The index closest to zero in the last 1/5 of each spectrum sample is determined.
    The index with the most repeats across the dataset estimated to be the
    reference wavelength.

    :param a_mts: TS-corrected absorption coefficient.
    :param wavelength_dim: The wavelength dimension name for a_mts.
    :return: The reference wavelength in nm.
    """
    wvl_idxs = xr.apply_ufunc(_estimate_reference_wavelength_index,
                              a_mts,
                              input_core_dims=[[wavelength_dim]],
                              vectorize=True)
    wvl_idxs = wvl_idxs.where(wvl_idxs != -999, drop=True)  # Drop bad estimations.
    idxs, counts = np.unique(wvl_idxs, return_counts=True)
    wvl_idx = idxs[np.nanargmax(counts)]
    if np.isnan(wvl_idx):
        raise ValueError('No reference wavelength found. Please check the input spectra.')
    wvl = float(a_mts[wavelength_dim].values[int(wvl_idx)])
    return wvl



"""
This module contains functions and classes for reading and parsing ACS data over RS232.
"""
from datetime import datetime, timezone
import serial

from .core import PACKET_REGISTRATION
from .structures import ACSPacket, ParsedPacket, DeviceCalibratedPacket, TSCorrectedPacket
from .packet import parse_packet, calibrate_packet, ts_correct_packet
from .dev import ACSDev


class ACSStream:
    BAUDRATE: int = 115200
    BYTESIZE: int = 8
    PARITY: str = 'N'
    STOPBITS: int = 1
    FLOWCONTROL: int = 0
    TIMEOUT: float = 1

    def __init__(self, port: str,
                 baudrate: int = BAUDRATE,
                 timeout: float = TIMEOUT) -> None:
        """
        Instantiate the ACSSerial class, reset the serial port, and clear any data in memory.

        :param port: The COM port the ACS is connected to.
        :param baudrate: The baudrate of the ACS. All ACS sensors come from the factory set to 115200 bps.
        :param timeout: The number of seconds to wait before actions timeout.
        :return: None
        """

        self.clear_packet_buffer()
        self.reset_serial(port=port, baudrate=baudrate, timeout=timeout)

    def clear_serial_buffers(self) -> None:
        """
        Clear the input and output serial buffers on the class assigned port.
        :return: None
        """
        self._serial.reset_input_buffer()
        self._serial.reset_output_buffer()

    def clear_packet_buffer(self):
        """
        Clear the packet buffer that is currently in memory.
        :return: None
        """
        self._buffer = bytearray()

    def reset_serial(self, port: str,
                     baudrate: int,
                     bytesize: int = BYTESIZE,
                     parity: str = PARITY,
                     stopbits: int = STOPBITS,
                     flowcontrol: int = FLOWCONTROL,
                     timeout: float = TIMEOUT):
        """
        Reset or instantiate serial object using the assigned settings.

        :param port: A string representing the COM port the ACS is connected to. e.g. 'COM1' or '/dev/ttyUSB0'
        :param baudrate: The baudrate the ACS communicates at. All ACS sensors come from the factory set to 115200 bps.
        :param bytesize: The number of bits in a byte. Default is 8.
        :param parity: The parity setting for the ACS. Default is 'N' for no parity.
        :param stopbits: The number of stop bits. Default is 1.
        :param flowcontrol: The flow control setting. Default is 0 for no flow control.
        :param timeout: The number of seconds to wait before timeout. Default is 1 second.
        :return: None
        """

        self._serial = serial.Serial()
        self._serial.port = port
        self._serial.baudrate = int(baudrate)
        self._serial.bytesize = int(bytesize)
        self._serial.parity = parity
        self._serial.stopbits = int(stopbits)
        self._serial.xonxoff = int(flowcontrol)
        self._serial.timeout = int(timeout)

    def connect(self) -> None:
        """
        Attempt to connect to the port assigned at class instantiation.

        :return: None
        """

        try:
            self._serial.open()
        except ConnectionError:
            raise serial.PortNotOpenError()

    def disconnect(self) -> None:
        """
        Disconnect from the port assigned at class instantiation.

        :return: None
        """

        self.clear_serial_buffers()
        self._serial.close()
        self.clear_packet_buffer()

    def read_stream(self) -> None:
        """
        Read the current number of bytes waiting in the serial port buffer and append them to a byte array in memory.

        :return: None
        """

        incoming = self._serial.read(self._serial.in_waiting)
        self._buffer.extend(incoming)

    def find_packet(self) -> ACSPacket:
        """
        Find the first full ACS packet within buffered serial data.

        :return: Raw ACS Packet data stored in a custom NamedTuple.
        """

        old_bytes, reg_bytes, remaining_bytes = self._buffer.partition(PACKET_REGISTRATION)
        daq_time = datetime.now(timezone.utc)
        if PACKET_REGISTRATION in remaining_bytes:
            packet_data, next_reg_bytes, next_loop_bytes = remaining_bytes.partition(PACKET_REGISTRATION)
            packet = reg_bytes + packet_data  # Rebuild the complete packet.
            self._buffer = next_reg_bytes + next_loop_bytes  # Redefine buffer.
            acs_packet = ACSPacket(daq_time=daq_time, full_packet=packet)
        else:
            acs_packet = ACSPacket(daq_time=daq_time, full_packet=None)
        return acs_packet

    def __enter__(self) -> object:
        """
        Context Manager: Automatically connect to the assigned port number when used in a "with" statement.

        :return: Class conditions.
        """

        self.connect()
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        """
        Context Manager: Exit and perform garbage cleanup.

        :param exc_type: The type of exception that was raised.
        :param exc_value: The value of the exception that was raised.
        :param traceback: The traceback of the exception that was raised.
        :return: None
        """

        self.disconnect()

    def parse_packet(self, acs_packet: ACSPacket) -> ParsedPacket:
        """ACSStream wrapper for the parse_packet function inherited from the packet module."""
        parsed_packet = parse_packet(acs_packet)
        return parsed_packet

    def calibrate_packet(self, parsed_packet, dev: ACSDev) -> DeviceCalibratedPacket:
        """ACSStream wrapper for the calibrate_packet function inherited from the packet module."""
        dev_cal_packet = calibrate_packet(parsed_packet, dev)
        return dev_cal_packet

    def ts_correct_packet(self, dev_cal_packet, temperature, salinity, dev: ACSDev) -> TSCorrectedPacket:
        """ACSStream wrapper for the ts_correct_packet function inherited from the packet module."""
        ts_corrected_packet = ts_correct_packet(dev_cal_packet, temperature, salinity, dev)
        return ts_corrected_packet



"""
This module contains functions that facilitate the estimation of other data products through the scientific work of
others.
"""

import xarray as xr


def compute_alh_676(a_p_650: xr.DataArray, a_p_676: xr.DataArray, a_p_715: xr.DataArray) -> xr.DataArray:
    """
    Compute absorption line height at 676 nm via Boss et al, 2007.
    https://link.springer.com/chapter/10.1007/978-1-4020-5824-0_9

    :param a_p_650: a_p data at 650 nm.
    :param a_p_676: a_p data at 676 nm.
    :param a_p_715: a_p data at 715 nm.
    :return: Absorption line height at 676 nm in m^-1.
    """

    alh = (a_p_676 - (39 / 65 * a_p_650) + (26 / 65 * a_p_715))

    if isinstance(alh, xr.DataArray):
        alh.attrs['ancillary_variables'] = [a_p_650.name, a_p_676.name, a_p_715.name]
        alh.attrs['long_name'] = 'Absorption Line Height at 676 nm'
        alh.attrs['units'] = 'm^-1'
        alh.attrs['reference'] = 'Boss et al. (2007)'
        alh.attrs['reference_doi'] = 'https://doi.org/10.1007/978-1-4020-5824-0_9'
    return alh


def estimate_chl(a_p_650: xr.DataArray,
                 a_p_676: xr.DataArray,
                 a_p_715: xr.DataArray,
                 alh_coeff: float = 0.0104) -> xr.DataArray:
    """
    Estimate chlorophyll-a from absorption line height via Roesler and Barnard 2013.
    https://doi.org/10.1016/j.mio.2013.12.003

    This function has only been tested on xarray.Datasets.

    :param a_p_650: a_p data at 650 nm.
    :param a_p_676: a_p data at 676 nm.
    :param a_p_715: a_p data at 715 nm.
    :param alh_coeff: (a_p line height coefficient value, default is 0.0104,
        which is the average from Table 1 in Roesler and Barnard 2013.
    :return: Chlorophyll-a concentration in mg/m^3.
    """

    abl = ((a_p_715 - a_p_650) / (715 - 650)) * (676 - 650) + a_p_650  # EQ 1 in Roesler and Barnard 2013
    alh = a_p_676 - abl  # EQ 2 in Roesler and Barnard 2013
    chl_alh = alh / alh_coeff  # EQ 3 in Roesler and Barnard 2013

    if isinstance(chl_alh, xr.DataArray):
        chl_alh.attrs['ancillary_variables'] = a_p_650.name
        chl_alh.attrs['alh_coeff'] = alh_coeff
        chl_alh.attrs['long_name'] = 'Chlorophyll-a Concentration from Absorption Line Height'
        chl_alh.attrs['units'] = 'mg/m^3'
        chl_alh.attrs['estimation_method'] = 'Roesler and Barnard (2013)'
        chl_alh.attrs['method_data_source'] = 'Cultures and Gulf of Maine'
        chl_alh.attrs['reference'] = 'Roesler and Barnard (2013)'
        chl_alh.attrs['reference_doi'] = 'https://doi.org/10.1016/j.mio.2013.12.003'
    return chl_alh


def estimate_poc(c_p_660: xr.DataArray, slope_offset: str | tuple | list) -> xr.DataArray:
    """
    Compute particulate organic carbon from particle beam attenuation at 660nm using a provided slope and offset.
    If the slope_offset value is a tuple or list, the first value is the slope and the second value is the offset.

    If the slope_offset value is a string, a predefined slope and offset will be used. Strings represent the first
    author and the year of the publication.

    Note: If supplying your own slope and offset, please ensure that the units are correct.

    :param c_p_660: particulate attenuation (c_p) at 660 nm.
    :param slope_offset: A string indicating the literary source for the slope and offset or
        a tuple or list containing the slope and offset values.
    :return: Estimated particulate organic carbon in mg/m3.
    """

    if isinstance(slope_offset, str):
        if slope_offset == 'gardner2006':  # Gardner et al. 2006, Global, All Seasons
            slope = 381
            offset = 9.4
            attrs = {'estimation_method': 'Gardner et al. (2006)',
                     'method_data_source': 'Global, All Seasons',
                     'reference': 'Gardner et al. (2006)',
                     'reference_doi': 'https://doi.org/10.1016/j.dsr2.2006.01.029'}
        elif slope_offset == 'stramski2008':  # Stramski et al. 2008, Fall, Pacific, Atlantic
            slope = 458
            offset = 10.7
            attrs = {'estimation_method': 'Stramski et al. (2008)',
                     'method_data_source': 'Fall, Pacific, Atlantic',
                     'reference': 'Stramski et al. (2008)',
                     'reference_doi': 'https://doi.org/10.5194/bg-5-171-2008'}
        elif slope_offset == 'behrenfeld2006':  # Behrenfeld and Boss 2006, Fall, Equatorial Pacific
            slope = 585
            offset = 7.6
            attrs = {'estimation_method': 'Behrenfeld and Boss (2006)',
                     'method_data_source': 'Fall, Equatorial Pacific',
                     'reference': 'Behrenfeld and Boss (2006)',
                     'reference_doi': 'https://doi.org/10.1357/002224006778189563'}
        elif slope_offset == 'cetenic2008':  # Cetinić et al. 2008, Spring, NE Atlantic
            slope = 391
            offset = -5.8
            attrs = {'estimation_method': 'Cetinić et al. (2008)',
                     'method_data_source': 'Spring, NE Atlantic',
                     'reference': 'Cetinić et al. (2008)',
                     'reference_doi': ' https://doi.org/10.1029/2011JC007771 '}
    elif isinstance(slope_offset, tuple | list):
        slope, offset = slope_offset

    poc = slope * c_p_660 + offset  # y = m * x + b

    if isinstance(poc, xr.DataArray):
        for k, v in attrs.items():
            poc.attrs[k] = v
        poc.attrs['ancillary_variables'] = c_p_660.name
        poc.attrs['slope'] = slope
        poc.attrs['offset'] = offset
        poc.attrs['long_name'] = 'Particulate Organic Carbon from Particle Beam Attenuation (c_p) at 660nm'
        poc.attrs['units'] = 'mg/m^3'
    return poc



from acspype.tscor import ACSTS4CorReader

TEST_FILEPATH = '../dev_tools/tscor/TS4.cor'


def test():
    tscor = ACSTS4CorReader(TEST_FILEPATH)

    assert len(tscor.wavelengths) == len(tscor.psi_t)
    assert len(tscor.wavelengths) == len(tscor.psi_s_c)
    assert len(tscor.wavelengths) == len(tscor.psi_s_a)



import numpy as np

def compute_base_uncertainty(accuracy: float, precision: float, other: float = 0.0) -> float:
    """Compute the base uncertainty associated with a sensor."""
    base_unc = np.sqrt(accuracy**2 + precision**2 + other**2)
    return base_unc






from datetime import datetime, timezone

from acspype import ACSDev
from acspype.packet import parse_packet, calibrate_packet, ts_correct_packet
from acspype import ACSPacket
from acspype.qaqc import FLAG, syntax_test

TEST_TIME = datetime.now(timezone.utc)

# TEST_PACKET is from ACS-00011 sampling in air.
TEST_PACKET = (b'\xff\x00\xff\x00\x02\xc0\x05\x01S\x00\x00\x0b\x01\xd0\x00\x00"\xadrc\xadA\x01\xe0\x02\xcc\x00H\x80'
               b'\xc3\x01T\x02\r\x01\x93\x01\xf4\x01\xc3\x02U\x01\xd1\x02<\x02\x07\x02\xa9\x02\x19\x02\x87\x02[\x03'
               b'\x0b\x02q\x02\xeb\x02\xbf\x03y\x02\xd6\x03[\x037\x03\xf7\x03C\x03\xdc\x03\xb9\x04s\x03\xb9\x04Y\x04G'
               b'\x05\n\x04F\x04\xf2\x04\xf3\x05\xb1\x04\xe1\x05\x9e\x05\xb4\x06n\x05\x98\x06`\x06\x99\x07G\x06Y\x07E'
               b'\x07\x8e\x08\x11\x07\x10\x08\x1f\x08z\x08\xe2\x07\xd7\x08\xf5\t}\t\xc4\x08\xb1\t\xe7\n\xa0\n\xc4\t\xa9'
               b'\n\xf1\x0b\xea\x0b\xce\n\xa3\x0c\x15\r:\x0c\xea\x0b\xb9\r;\x0e\xb7\x0e\x1e\x0c\xd4\x0e\x8f\x10=\x0f['
               b'\x0e\r\x0f\xd8\x11\xf2\x10\xd0\x0fu\x11j\x13\xeb\x12`\x10\xec\x13\x1b\x15\xfc\x13\xe8\x12\\\x14\xcc'
               b'\x18\x0e\x15\x8b\x14\x00\x16\x84\x1ao\x17E\x15\x92\x18y\x1c\xbb\x18\xee\x17+\x1aH\x1f\x19\x1a\xa9'
               b'\x18\xcd\x1c-!\x86\x1c?\x1aI\x1d\xf9#\xc8\x1d\xe7\x1b\xfe\x1f\xb8&b\x1f\xd1\x1d\xe0!\xd9)A!\xab\x1f'
               b'\xa2#\xf5+\xf6#f!e%\xd8.\xbb%b#o\'\xfc1\xeb\'~%\x8d*^5;)\xde(\x1c,\xdf98,\x9a*\xb6/\xfe=M/\n-$2'
               b'\xb6A01\x8f/\xa75|E23\xd11\xea8\x0eH\xde604v:\x9fM\x068\xa96\xdf=}P\xf3:\xa28\xc3?\xc3T+<\x99:\xebA'
               b'\xddW\xcf>\xa8<\xf3DF[9@\x98?\x12FY^\xe0C\x16A\xa4I\x1ec6E"CQK\xbef\'J\xb5H\x1dQ\x92m\xbbL~I\xf4S'
               b'\x92q\x10N.K\xb7U\x96tLO\x85M(W\x18w\x0bP\xd5N\x8eX\xa0y\xbcQ\xf8O\xb8Y\xf9|\x18S*Q\x07[>~\xaaTQRF'
               b'\\\xcc\x81%T\xacR\xba]P\x82kT\x9fR\xd7]K\x83\x1fT\x7fR\xca]D\x83\x97T+R\x85\\\xfa\x83\xbeS\xa1R\x04'
               b'\\x\x83|R\xe2QU[\xc2\x82\xe9Q\x81O\xfdZv\x81FO\x96N\x11XZ~\xb1MpK\xfdV!{\xcbJ\xe9I\xa2SCxrI\x8cHjQ'
               b'\x94v\xdfH\xa5GrQ\x07u\x9cFPE\x16N\x8er\x1dC\xceB\x9dK\xa5naA[@-I\x07j\xac>\x90=mE\xeefi;\xef:\xd9B'
               b'\xf3b\\9\\8G@+^I6\x825b=\x17Y\xa83D2\x199\x83T\\0\x02.\xe75\xc6O&-T,;2\xc3J\xd0*\xde)\xba0\x1cF'
               b'\xb7()&\xfc-\x1cB5%r$D*\x03=\xb6"\xe7!\xb3\'19x \\\x1f2$\\5E\x1e\x18\x1c\xf1!\xc81\x82\x1c\x07\x1a'
               b'\xd8\x1f\x8b.\x03\x19\xe8\x18\xb6\x1d\x1b*r\xf8\xad\x00')

TEST_TEMPERATURE = 12.345  # Dummy temperature.
TEST_SALINITY = 34.567  # Dummy salinity.


def test():
    dev = ACSDev('../dev_tools/test_files/ACS-00011_2022-10-20.dev')

    acs_packet = ACSPacket(daq_time=TEST_TIME, full_packet=TEST_PACKET)
    assert syntax_test(acs_packet.full_packet) == FLAG.OK

    parsed_packet = parse_packet(acs_packet)

    calibrated_packet = calibrate_packet(parsed_packet, dev)
    assert dev.serial_number == calibrated_packet.serial_number

    ts_corrected_packet = ts_correct_packet(calibrated_packet, TEST_TEMPERATURE, TEST_SALINITY, dev)

    assert ts_corrected_packet.temperature == TEST_TEMPERATURE



"""This module contains core constants and functions used across acspype."""

from math import nan, sqrt
from numpy.typing import NDArray
import uncertainties
import xarray as xr

NUM_PAT = r"[+-]?[0-9]*[.]?[0-9]+"  # REGEX for any number, float or int, positive or negative.

# ---------- PACKET HANDLING ---------- #
PACKET_REGISTRATION = b"\xff\x00\xff\x00"  # Start of every ACS packet.
PAD_BYTE = b"\x00"  # End of every ACS packet.
WVL_BYTE_OFFSET = 4 + 2 + 1 + 1 + 1 + 3 + 2 + 2 + 2 + 2 + 2 + 2 + 2 + 4 + 1
NUM_CHECKSUM_BYTES = 2  # Number of bytes in the checksum.
PACKET_HEAD = "!4cHBBl7HIBB"
PACKET_TAIL = "Hx"  # struct descriptor for the last 3 bytes of an ACS packet (checksum + pad byte).
LPR = len(PACKET_REGISTRATION)

# ---------- FILE CREATION ---------- #
NC_ENCODING = {
    "time": {"units": "milliseconds since 1970-01-01"}
}  # Recommended xr.Dataset to netcdf encoding for time.

# ---------- PHYSICAL QUANTITIES ---------- #
EST_FLOW_CELL_VOLUME = 30  # in mL, from the ACS Protocol Document, Rev Q.


# ---------- INSTRUMENT SPECIFICATIONS ---------- #
class SPECS:
    class ACS:
        ACCURACY: float = 0.01
        PRECISION_SHORT: float = 0.012  # Max between 400-449nm
        PRECISION_LONG: float = 0.003  # Max between 450-730nm
        UNC_SHORT: float = sqrt(
            (0.012**2) + (0.01**2)
        )  # Combined uncertainty from base precision/accuracy following Eq. 3b from Csavina et al. 2017.
        UNC_LONG: float = sqrt(
            (0.003**2) + (0.01**2)
        )  # Combined uncertainty from base precision/accuracy following Eq. 3b from Csavina et al. 2017.
        GROSS_MIN: float = 0.001
        GROSS_MAX: float = 10.0

    class INTERNAL_TEMPERATURE:
        ACCURACY: float = 0.1  # From email with Sea-Bird Tech Support.
        UNC: float = 0.1

    class EXTERNAL_TEMPERATURE:
        ACCURACY: float = nan
        UNC: float = nan

    class PRESSURE:
        ACCURACY: float = nan  # Unknown
        UNC: float = nan


def is_uncertainties(values: NDArray, max_checks: int = 5) -> bool:
    """
    Check if an object might be from the uncertainties package.
    :param values: The input to check.
    :param max_checks: The number of times to check in case the object is a nested NDArray.
    :return: A boolean indicating whether or not the object is from the uncertainties package.
    """
    if isinstance(values, xr.DataArray):
        values = values.values
    if isinstance(values, uncertainties.core.Variable) or isinstance(values, uncertainties.unumpy.core.matrix):
        return True
    else:
        for i in range(max_checks):
            try:
                values = values[0]
            except:
                break
        if isinstance(values, uncertainties.core.Variable):
            return True
        else:
            return False



"""
This module contains functions for ACS data quality assurance and quality control (QAQC).
QAQC tests use QARTOD-style flagging and meanings, unless an alternative is specified.
"""

from datetime import datetime, timedelta
import numpy as np
from numpy.typing import NDArray
from struct import calcsize
import xarray as xr

from .core import (PACKET_HEAD, PACKET_TAIL, LPR, PACKET_REGISTRATION, PAD_BYTE)
from .dev import ACSDev
from .packet import unpack_packet


class FLAG:
    """
    Flag values for the ACS QAQC tests. The following flags follow the QARTOD flag meanings.

    :param OK: Indicates that the data passed the test.
    :param PASS: Indicates that the data passed the test.
    :param NOT_EVALUATED: Indicates that the data was not evaluated. Appearance of this flag indicates a problem with
        the test function and not the data. It defeats the purpose of running a QAQC if the data is not evaluated.
    :param SUSPECT: Indicates that the data is suspect.
        This flag indicates that the data should be reviewed more closely.
    :param HIGH_INTEREST: Indicates that the data is of high interest. Although it uses the same flag value as SUSPECT,
        the context of the test and environmental conditions should be considered when reviewing the data.
    :param FAIL: Indicates that the data failed the test conditions.
    :param BAD: Indicates that the data failed the test conditions.
    :param MISSING_DATA: Indicates that the input data or ancillary data is missing (NaN), resulting in data that
        cannot be used.
    """

    OK: int = 1
    PASS: int = 1
    NOT_EVALUATED: int = 2
    SUSPECT: int = 3
    HIGH_INTEREST: int = 3
    FAIL: int = 4
    BAD: int = 4
    MISSING_DATA: int = 9


def gap_test(now: datetime,
             time_stmp: datetime,
             record_length: int,
             buffer_length: int | None,
             time_inc: float = 0.25) -> int:
    """
    Assess the gap between the current time and the timestamp of the packet.
    This is a modified form of the generic QARTOD gap test.

    :param now: The time at which the test is being run. Must be compatibile with a datetime.datetime object.
    :param time_stmp: The timestamp of the packet. Must be compatible with a datetime.datetime object.
    :param buffer_length: The number of bytes in the serial buffer. If None, the corresponding section of the test is
        skipped. Determined through the protected attribute "_buffer" in the ACSStream class.
    :param record_length: The record length of the packet. This is the expected number of bytes in the packet that is
        within each ACS packet header.
    :param time_inc: The time increment to use for assessing timestamp gaps. Default is 0.25 seconds, which is the
        approximate time that it takes the ACS to send a complete packet.
    :return: A flag indicating pass or fail.
    """
    if now - time_stmp > timedelta(seconds=time_inc):  # Defined in QARTOD Ocean Optics Manual.
        return FLAG.FAIL
    elif buffer_length is not None:
        if buffer_length > record_length:
            """
            This is a custom take on the gap test. 
            If for some reason the buffer length exceeds the previous frame length, that would indicate that
            the buffer is filling up faster than the packets can be unpacked. This could ultimately result in the 
            timestamp of the value being off by one or multiples of 250ms periods, depending on the buffer length. 
            This could also indicate an issue with the timing of a data acquisition thread.
            """
            return FLAG.FAIL
    else:
        return FLAG.PASS


def syntax_test(full_packet: bytearray) -> int:
    """
    Test the syntax of an ACS packet.
    This test expands on the QARTOD syntax test.
    Failure of this test indicates that the packet is malformed or incomplete.

    :param full_packet: A full packet from the ACS, including registration bytes and the pad byte.
        See acspype.core for the static syntax of the header and tail of each packet. The unpack_packet function in the
        acspype.packet module shows how to dynamically unpack the packet using the packet header.
    :return: Flag indicating pass or fail.
    """

    raw, checksum = unpack_packet(full_packet)
    remaining_packet_size = int(raw[LPR + 13] * 2 * 2)
    full_packet_descriptor = PACKET_HEAD + f"{remaining_packet_size}H" + PACKET_TAIL
    expected_length = raw[LPR + 0]
    if full_packet[:len(PACKET_REGISTRATION)] != PACKET_REGISTRATION:  # If no packet registration...
        return FLAG.FAIL
    elif full_packet[-1] != int.from_bytes(PAD_BYTE):  # If no pad byte...
        return FLAG.FAIL
    elif calcsize(full_packet_descriptor) != len(full_packet):  # If the descriptor doesn't match the packet length...
        return FLAG.FAIL
    elif len(full_packet) - calcsize(PACKET_TAIL) != expected_length:  # If the expected record length doesn't match...
        return FLAG.FAIL
    elif sum(full_packet[:-calcsize(PACKET_TAIL)]) >= 65536:  # If the checksum doesn't add up....
        if checksum != sum(full_packet[:-calcsize(PACKET_TAIL)]) % 65536:
            return FLAG.FAIL
        else:
            return FLAG.PASS
    elif sum(full_packet[:-calcsize(PACKET_TAIL)]) < 65536:
        if checksum != np.uint16(sum(full_packet[:-calcsize(PACKET_TAIL)])):
            return FLAG.FAIL
        else:
            return FLAG.PASS
    else:
        return FLAG.PASS


def elapsed_time_test(elapsed_time: int | xr.DataArray, fail_threshold: int = 1000 * 60,
                      suspect_threshold: int = 1000 * 60 * 3) -> int | xr.DataArray:
    """
    Assess the elapsed time since the instrument was turned on. The elapsed time is always output from the ACS as the
    number of milliseconds since the instrument first received power.
    This is a custom test that is based on information in the ACS manual.
    Generally, the ACS takes 5-10 minutes to warm-up. Waiting this long may not be practical in some situations.
    It is recommended to flag the first minute of data as poor quality and then minutes 2-3 as suspect in applications
    where the ACS is on for less than 10 minutes at a time.

    :param elapsed_time: The elapsed time parsed from an ACS packet.
        Represents the number of milliseconds that have passed since the ACS was turned on.
    :param fail_threshold: The amount of time, from initial instrument power on in milliseconds, where data is
        considered to be of poor quality.
    :param suspect_threshold: The amount of time in milliseconds between the fail and
        suspect threshold where data is considered to be suspect.
    :return: Flag indicating pass, suspect, or fail.
    """

    if not isinstance(elapsed_time, xr.DataArray):
        if elapsed_time <= fail_threshold:
            return FLAG.FAIL
        elif fail_threshold < elapsed_time <= suspect_threshold:
            return FLAG.SUSPECT
        else:
            return FLAG.PASS
    else:
        flags = xr.full_like(elapsed_time, FLAG.NOT_EVALUATED)
        flags = xr.where(elapsed_time <= fail_threshold, FLAG.FAIL, flags)
        flags = xr.where((elapsed_time > fail_threshold) & (elapsed_time <= suspect_threshold), FLAG.SUSPECT, flags)
        flags = xr.where(elapsed_time > suspect_threshold, FLAG.PASS, flags)

        # Assign attributes to the flags if an xarray.DataArray.
        flags.attrs['ancillary_variables'] = elapsed_time.name
        flags.attrs['fail_threshold'] = fail_threshold
        flags.attrs['suspect_threshold'] = suspect_threshold
        flags.attrs['threshold_units'] = 'milliseconds'
        flags.attrs['test_name'] = 'elapsed_time_test'
        return flags


def internal_temperature_test(internal_temperature: float | xr.DataArray,
                              dev: ACSDev) -> float | xr.DataArray:
    """
    Assess the internal temperature of the ACS. This is a custom test that is based on information within the manual
    and the device file. The ACS is calibrated at specific temperature bins, which are listed in the device file.
    Typically, the internal temperature correction is linearly interpolated from the delta T values in the device file.
    If the internal temperature of the sensor exceeds the range of those temperature bins,
    then data are considered to be suspect.

    :param internal_temperature: The internal temperature of the ACS.
    :param dev: The device file for the ACS as an ACSDev object.
    :return: Flag indicating pass or suspect.
    """

    min_t = min(dev.tbin)
    max_t = max(dev.tbin)
    if not isinstance(internal_temperature, xr.DataArray):
        if min_t <= internal_temperature <= max_t:
            return FLAG.PASS
        else:
            return FLAG.SUSPECT
    else:
        flags = xr.full_like(internal_temperature, FLAG.NOT_EVALUATED).astype(int)
        flags = xr.where((min_t > internal_temperature) & (max_t < internal_temperature), FLAG.SUSPECT, flags)
        flags = xr.where((min_t <= internal_temperature) & (max_t >= internal_temperature), FLAG.PASS, flags)

        # Assign attributes to the flags if an xarray.DataArray.
        flags.attrs['ancillary_variables'] = internal_temperature.name
        flags.attrs['minimum_temperature_bin'] = min_t
        flags.attrs['maximum_temperature_bin'] = max_t
        flags.attrs['temperature_bin_units'] = 'degrees Celsius'
        flags.attrs['test_name'] = 'internal_temperature_test'
        return flags


def inf_nan_test(uncorrected: NDArray[float] | xr.DataArray) -> int | xr.DataArray:
    """
    Assess if the uncorrected data contains any NaN or Inf values. This is a custom test that can be used to identify
    instances where the reference counts are zero, which produces mathematical errors in the conversion to geophysical
    units. Zero reference counts happen occasionally and can probably be discarded during analysis. Repeated samples
    containing zero reference counts may indicate a problem with the instrument.

    :param uncorrected: Uncorrected a or c data.
    :return: Flag indicating pass or fail.
    """
    if not isinstance(uncorrected, xr.DataArray):
        if np.any(np.isinf(uncorrected)) or np.any(np.isnan(uncorrected)):
            return FLAG.FAIL
        else:
            return FLAG.PASS
    else:
        flags = xr.full_like(uncorrected.time.astype(int), FLAG.PASS).astype(int)
        flags = xr.where((np.any(np.isinf(uncorrected), axis=1)), FLAG.FAIL, flags)
        flags = xr.where((np.any(np.isnan(uncorrected), axis=1)), FLAG.FAIL, flags)

        # Assign attributes to the flags if an xarray.DataArray.
        flags.attrs['ancillary_variables'] = uncorrected.name
        flags.attrs['test_name'] = 'inf_nan_test'
        return flags


def gross_range_test(mts: xr.DataArray,
                     sensor_min: float = -0.005, sensor_max: float = 12.00,
                     user_min: float = 0.001, user_max: float = 10) -> xr.DataArray:
    """
    Assess if the TS-corrected data are within the limitations of the instrument. From the ACS manual (Rev I),
    values of 12 m^-1 may occur at shorter wavelengths. The valid dynamic range of the sensor is 0-10 m^-1.

    The return is a set of flags in the same shape as the input data. Note that fail flags may occur in the
    red wavelengths of absorption spectrum consecutively toward the end of each spectrum, which does not necessarily
    make the entire spectrum of poor quality. Users should consider the pass/fail state of the rest of the spectrum
    before making a decision on the validity of the data.

    :param mts: TS-corrected absorption or attenuation.
    :param sensor_min: The minimum sensor range. Default is -0.005, which is effectively 0 as described in the manual.
    :param sensor_max: The maximum sensor range. Default is 12.00, which is a value that is mentioned in Rev I of the user's guide.
    :param user_min: Operator set minimum for flagging suspect data. Default is 0.001 which is the minimum of the dynamic range.
    :param user_max: Operator set maximum for flagging suspect data. Default is 10 which is the maximum of the dynamic range.
    :return: Flag indicating pass, fail, or suspect.
    """

    if not isinstance(mts, xr.DataArray):
        raise NotImplementedError("Functionality for singletons not yet implemented.")

    else:
        flags = xr.full_like(mts, FLAG.NOT_EVALUATED).astype(int)
        flags = flags.where((mts > sensor_min) & (mts < sensor_max), FLAG.FAIL)
        flags = flags.where((mts > user_min) | (mts < sensor_min), FLAG.SUSPECT)
        flags = flags.where((mts < user_max) | (mts > sensor_max), FLAG.SUSPECT)
        flags = flags.where((mts <= user_min) | (mts >= user_max), FLAG.PASS)

        # Assign attributes to the flags if an xarray.DataArray.
        flags.attrs['ancillary_variables'] = mts.name
        flags.attrs['sensor_min'] = sensor_min
        flags.attrs['sensor_max'] = sensor_max
        flags.attrs['operator_min'] = user_min
        flags.attrs['operator_max'] = user_max
        flags.attrs['threshold_units'] = 'm^-1'
        flags.attrs['test_name'] = 'gross_range_test'
        return flags


def discontinuity_offset_test(discontinuity_offset: xr.DataArray,
                              median_multiplier: float = 3,
                              fail_threshold: float = 10) -> xr.DataArray:
    """
    Flag a discontinuity offset value as pass or suspect. This is a custom test that uses a multiple of the median
    of the offset to assess if the offset is acceptable.
    Spectra with significantly large discontinuity offsets are likely to be of poor quality, but should be assessed
    using other QAQC tests before being discarded, such as the a_gt_c_test, which may likely confirm the quality of
    the data.

    This function is only intended for use with xarray.DataArray objects.

    :param discontinuity_offset: The discontinuity offset value for absorption or attenuation.
    :param median_multiplier: The multiplier of the median. Values within the multiplier range are deemed acceptable.
    :param fail_threshold: The threshold for flagging the offset as fail.
        Default is 10, or the maximum value of the ACS range.
    :return: The flag of the discontinuity offset, which maintains the same size as the time dimension.
    """

    flags = xr.full_like(discontinuity_offset, FLAG.PASS).astype(int)
    _median = discontinuity_offset.median(skipna=True)  # Use the timeseries median to assess the offset.
    flags = flags.where((np.abs(discontinuity_offset) < np.abs(_median) * median_multiplier), FLAG.SUSPECT)
    flags = flags.where((np.abs(discontinuity_offset) < fail_threshold), FLAG.FAIL)

    # Assign attributes to the flags if an xarray.DataArray.
    flags.attrs['ancillary_variables'] = discontinuity_offset.name
    flags.attrs['median_multiplier'] = median_multiplier
    flags.attrs['fail_threshold'] = fail_threshold
    flags.attrs['suspect_threshold'] = np.abs(_median) * median_multiplier
    flags.attrs['threshold_units'] = 'm^-1'
    flags.attrs['test_name'] = 'discontinuity_offset_test'
    return flags


def blanket_gross_range_test(nd_gross_range_results: xr.DataArray,
                             wavelength_dim: str,
                             suspect_threshold: float,
                             fail_threshold: float,
                             include_suspect_flags: bool = False) -> xr.DataArray:
    """
    Assess the validity of the gross range test results as a blanket flag for the spectra.
    This is an experimental test.
    The test considers the number of wavelength bins that were previously flagged as suspect or fail for exceeding the
    gross range limits. If X percent of wavelength bins were flagged as fail (or fail/suspect) and exceed the
    user-defined fail threshold, then the entire spectrum is flagged as fail. If the percentage is between the suspect
    and fail thresholds then the spectra is flagged as suspect.
    The threshold represents the percentage threshold between 0-1 (0-100%).

    :param nd_gross_range_results: The N-dimensional results from the gross range test. Effectively a single flag for
        every value along the time/wavelength dimension.
    :param wavelength_dim: The wavelength dimension of the results.
    :param suspect_threshold: The percentage threshold for suspect data.
    :param fail_threshold: The percentage threshold for fail data.
    :param include_suspect_flags: If True, include suspect flags along the spectra in the total fail.
    :return: A blanket flag for the spectra, which maintains the same shape as the time dimension.
    """
    if fail_threshold < suspect_threshold:
        raise ValueError('fail_threshold must be greater than suspect_threshold.')

    num_wvls = len(nd_gross_range_results[wavelength_dim])

    if include_suspect_flags is True:
        _flags_bool = xr.where((nd_gross_range_results == FLAG.BAD) | (nd_gross_range_results == FLAG.SUSPECT), 1, 0)
    else:
        _flags_bool = xr.where(nd_gross_range_results == FLAG.BAD, 1, 0)

    _flags_bool_sum = np.sum(_flags_bool, axis=1)
    ratio = _flags_bool_sum / num_wvls
    flags = xr.full_like(nd_gross_range_results.time.astype(int), FLAG.NOT_EVALUATED).astype(int)

    flags = flags.where((ratio < suspect_threshold) & (ratio > fail_threshold), FLAG.SUSPECT)
    flags = flags.where(ratio < fail_threshold, FLAG.FAIL)

    flags = flags.where(ratio >= suspect_threshold, FLAG.PASS)

    # Assign attributes to the flags if an xarray.DataArray.
    flags.attrs['ancillary_variables'] = nd_gross_range_results.name
    flags.attrs['suspect_threshold'] = suspect_threshold
    flags.attrs['fail_threshold'] = fail_threshold
    flags.attrs['threshold_units'] = 'm^-1'
    flags.attrs['includes_suspect_flags'] = include_suspect_flags
    flags.attrs['test_name'] = 'blanket_gross_range_test'
    return flags


def a_gt_c_test(absorption: xr.DataArray, attenuation: xr.DataArray) -> xr.DataArray:
    """
    Assess if the absorption is greater than the attenuation. Having absorption (a) greater than attenuation (c)
    is (mostly) impossible. This test can be run on uncorrected data to help assess sensor drift.
    It is also recommended that this test be run on scattering corrected absorption with ts-corrected
    attenuation. This test is not related to QARTOD, but is a custom test based on reality checks in the
    ACS protocol document.
    This test is only intended for use with xarray.DataArray objects.

    :param absorption: Absorption data. Either uncorrected a or a_mts_scatcorr.
    :param attenuation: Attenuation data. Either uncorrected c or ts corrected c.
    :return: Flag indicating if absorption is greater than attenuation.
    """

    flags = xr.full_like(absorption, FLAG.NOT_EVALUATED).astype(int)
    flags = flags.where(absorption <= attenuation, FLAG.SUSPECT)
    flags = flags.where(absorption > attenuation, FLAG.PASS)

    # Assign attributes to the flags if an xarray.DataArray.
    flags.attrs['ancillary_variables'] = [absorption.name, attenuation.name]
    flags.attrs['test_name'] = 'a_gt_c_test'
    return flags


def rolling_variance_test(mts: xr.DataArray,
                          use_mean: str = 'rolling',
                          window_size: int = 4 * 60 + 1,
                          exceedance_threshold: float = 0.25,
                          min_periods: int | None = 1) -> xr.DataArray:
    """
    Apply a rolling variance test to a measured absorption or attenuation coefficient.

    :param mts: The measured absorption or attenuation coefficient. For absorption the recommendation is to use
        a scattering corrected measurement (e.g. a_mts_proportional).
        For attenuation the recommendation is to use the TS-corrected measurement.
    :param use_mean: Indicates whether to use the timeseries mean or a rolling window mean along the time dimension.
        If 'rolling', the rolling mean at each wavelength bin along the window size is used.
        If 'timeseries', the timeseries mean at each wavelength bin is used.
    :param window_size: The centered window size for the variance window
        and for the mean window, if use_mean = 'timeseries'.
    :param exceedance_threshold: The percentage of the mean that the variance must exceed to be flagged as suspect.
        Represented as a value between 0-1 (0-100%).
    :param min_periods: The minimum number of periods to apply this test. Only applies if use_mean = 'rolling'.
        Same functionality as the min_periods argument for xarray.DataArray.rolling. The default is 1, which generally
        means that this test is not applied to the first window in the timeseries.
    :return: A flag indicating SUSPECT or PASS.
    """

    use_mean = use_mean.lower()
    if use_mean == 'timeseries':
        m_mean = mts.mean(dim='time', skipna=True)
        mean_type = 'timeseries_mean'
    elif use_mean == 'rolling':
        m_mean = mts.rolling(time=window_size, center=True, min_periods=min_periods).mean(skipna=True)
        mean_type = 'rolling_mean'
    else:
        raise ValueError("use_mean must be 'timeseries' or 'rolling'.")

    m_var = mts.rolling(time=window_size, center=True, min_periods=min_periods).var(skipna=True)  # Rolling variance.

    flags = xr.full_like(mts, FLAG.NOT_EVALUATED).astype(int)
    flags = flags.where(m_var < exceedance_threshold * m_mean, FLAG.SUSPECT)
    flags = flags.where(m_var >= exceedance_threshold * m_mean, FLAG.PASS)

    # Assign attributes to the flags if an xarray.DataArray.
    flags.attrs['ancillary_variables'] = mts.name
    flags.attrs['rolling_variance_test_window_size'] = window_size
    flags.attrs['exceedance_threshold'] = exceedance_threshold
    flags.attrs['mean_type'] = mean_type
    flags.attrs['min_periods'] = min_periods
    flags.attrs['test_name'] = 'rolling_variance_test'
    return flags



"""
This module contains NamedTuple structures that ACS data can be dumped into for easy tracking and exporting.
Each NamedTuple is assigned methods that allows for the export of data to a dictionary or to a formatted
xarray.Dataset.
"""

from datetime import datetime
import numpy as np
from numpy.typing import ArrayLike
from typing import NamedTuple
import xarray as xr


class ACSPacket(NamedTuple):
    """
    A class for representing a time-stamped ACS packet.

    :param daq_time: The host computer time the packet was acquired.
    :param full_packet: The full packet, including the registration bytes and the checksum + pad bytes.
    """
    daq_time: datetime
    full_packet: bytes | bytearray | None

    def to_dict(self) -> dict:
        """
        Export the ACSPacket as a dictionary.

        :return: A dict representation of the ACSPacket, where keys are attributes and values are attribute values.
        """
        return self._asdict()

    def to_xarray(self) -> xr.Dataset:
        """
        Export the ACSPacket as an xarray.Dataset.

        :return: An xarray.Dataset representation of the ACSPacket, where coordinates are the daq_time
            and the data variable is the full_packet.
        """
        ds = xr.Dataset()
        ds = ds.assign_coords({'daq_time': self.daq_time})
        ds['full_packet'] = bytes(self.full_packet)
        return ds


class ParsedPacket(NamedTuple):
    """
    A class for representing a parsed ACS packet.

    :param daq_time: The host computer time the packet was acquired, usually passed from an ACSPacket.
    :param record_length: The length of the packet in bytes.
    :param packet_type: The type of packet.
    :param sn_int: The serial number of the device represented as an integer.
    :param a_reference_dark: The dark reference counts for the absorption channel.
    :param a_signal_dark: The dark signal counts for the absorption channel.
    :param raw_external_temperature: The raw external temperature reading in counts.
    :param raw_internal_temperature: The raw internal temperature reading in counts.
    :param c_reference_dark: The dark reference counts for the attenuation channel.
    :param c_signal_dark: The dark signal counts for the attenuation channel.
    :param elapsed_time: The elapsed time since power was supplied to the ACS, in milliseconds.
    :param number_of_wavelengths: The number of wavelengths for each channel.
    :param c_reference: The attenuation reference in counts.
    :param a_reference: The absorption reference in counts.
    :param c_signal: The attenuation signal in counts.
    :param a_signal: The absorption signal in counts.
    :param checksum: The checksum of the packet.
    """
    daq_time: datetime
    record_length: int
    packet_type: int
    # reserved_1: int  # Commented out for unlikely reimplementation in the future.
    sn_int: int
    a_reference_dark: int
    # raw_pressure: int
    a_signal_dark: int
    raw_external_temperature: int
    raw_internal_temperature: int
    c_reference_dark: int
    c_signal_dark: int
    elapsed_time: int
    # reserved_2: int
    number_of_wavelengths: int
    c_reference: tuple[int, ...]
    a_reference: tuple[int, ...]
    c_signal: tuple[int, ...]
    a_signal: tuple[int, ...]
    checksum: int

    def to_dict(self) -> dict:
        """
        Export the ParsedPacket as a dictionary.

        :return: A dict representation of the ParsedPacket, where keys are attributes and values are attribute values.
        """
        return self._asdict()

    def to_xarray(self) -> xr.Dataset:
        """
        Export the ParsedPacket as an xr.Dataset.

        :return: An xarray.Dataset representation of the ParsedPacket.
        """
        ds = xr.Dataset()
        ds = ds.assign_coords({'daq_time': [self.daq_time],
                               'wavelength_index': list(range(self.number_of_wavelengths))})
        ds['record_length'] = self.record_length
        ds['packet_type'] = self.packet_type
        # ds['reserved_1'] = self.reserved_1
        ds['sn_int'] = self.sn_int
        ds['a_reference_dark'] = self.a_reference_dark
        # ds['raw_pressure'] = self.raw_pressure
        ds['a_signal_dark'] = self.a_signal_dark
        ds['raw_external_temperature'] = self.raw_external_temperature
        ds['raw_internal_temperature'] = self.raw_internal_temperature
        ds['c_reference_dark'] = self.c_reference_dark
        ds['c_signal_dark'] = self.c_signal_dark
        ds['elapsed_time'] = self.elapsed_time
        # ds['reserved_2'] = self.reserved_2
        ds['number_of_wavelengths'] = self.number_of_wavelengths
        ds['c_reference'] = (['daq_time', 'wavelength_index'], [self.c_reference])
        ds['a_reference'] = (['daq_time', 'wavelength_index'], [self.a_reference])
        ds['c_signal'] = (['daq_time', 'wavelength_index'], [self.c_signal])
        ds['a_signal'] = (['daq_time', 'wavelength_index'], [self.a_signal])
        ds['checksum'] = self.checksum
        return ds


class DeviceCalibratedPacket(NamedTuple):
    """
    A class for representing a device file-calibrated ACS packet.

    :param daq_time: The host computer time the packet was acquired, usually passed from an ParsedPacket.
    :param a_wavelength: The wavelengths for the absorption channel.
    :param c_wavelength: The wavelengths for the attenuation channel.
    :param sn_hexdec: The serial number of the device represented as a hexadecimal string.
    :param serial_number: The serial number of the device represented as a string.
    :param internal_temperature: The internal temperature of the device in degrees Celsius.
    :param external_temperature: The external temperature of the device in degrees Celsius.
    :param a_uncorrected: The uncorrected absorption signal in inverse meters.
    :param c_uncorrected: The uncorrected attenuation signal in inverse meters.
    :param discontinuity_index: The index of the discontinuity in both channels.
    :param a_discontinuity_offset: The offset for the absorption channel discontinuity.
    :param c_discontinuity_offset: The offset for the attenuation channel discontinuity.
    :param a_m_discontinuity: The absorption coefficient without discontinuity correction applied.
    :param c_m_discontinuity: The attenuation coefficient without discontinuity correction applied.
    :param a_m: The measured absorption coefficient in inverse meters.
    :param c_m: The measured attenuation coefficient in inverse meters.
    """

    daq_time: datetime
    a_wavelength: tuple[float, ...]
    c_wavelength: tuple[float, ...]
    sn_hexdec: str
    serial_number: str
    internal_temperature: float
    external_temperature: float
    a_uncorrected: ArrayLike
    c_uncorrected: ArrayLike
    discontinuity_index: int
    a_discontinuity_offset: float
    c_discontinuity_offset: float
    a_m_discontinuity: ArrayLike
    c_m_discontinuity: ArrayLike
    a_m: ArrayLike
    c_m: ArrayLike

    def to_dict(self) -> dict:
        """
        Export the DeviceCalibratedPacket as a dictionary.

        :return: A dict representation of the DeviceCalibratedPacket, where keys are attrs and values are attr values.
        """
        return self._asdict()

    def to_xarray(self) -> xr.Dataset:
        """
        Export the DeviceCalibratedPacket as an xarray.Dataset.

        :return: An xarray.Dataset representation of the DeviceCalibratedPacket.
        """
        ds = xr.Dataset()
        ds = ds.assign_coords({'daq_time': [self.daq_time],
                               'a_wavelength': np.array(self.a_wavelength),
                               'c_wavelength': np.array(self.c_wavelength)})
        ds['sn_hexdec'] = self.sn_hexdec
        ds['serial_number'] = self.serial_number
        ds['internal_temperature'] = self.internal_temperature
        ds['external_temperature'] = self.external_temperature
        ds['a_uncorrected'] = (['daq_time', 'a_wavelength'], [self.a_uncorrected])
        ds['c_uncorrected'] = (['daq_time', 'c_wavelength'], [self.c_uncorrected])
        ds['discontinuity_index'] = self.discontinuity_index
        ds['a_discontinuity_offset'] = self.a_discontinuity_offset
        ds['c_discontinuity_offset'] = self.c_discontinuity_offset
        ds['a_m_discontinuity'] = (['daq_time', 'a_wavelength'], [self.a_m_discontinuity])
        ds['c_m_discontinuity'] = (['daq_time', 'c_wavelength'], [self.c_m_discontinuity])
        ds['a_m'] = (['daq_time', 'a_wavelength'], [self.a_m])
        ds['c_m'] = (['daq_time', 'c_wavelength'], [self.c_m])
        return ds


class TSCorrectedPacket(NamedTuple):
    """
    A class for representing a temperature and salinity-corrected ACS packet.

    :param daq_time: The host computer time the packet was acquired, usually passed from a DeviceCalibratedPacket.
    :param a_wavelength: The wavelengths for the absorption channel.
    :param c_wavelength: The wavelengths for the attenuation channel.
    :param temperature: The temperature of the water in degrees Celsius.
    :param salinity: The salinity of the water in practical salinity units (PSU).
    :param a_mts: The temperature and salinity-corrected absorption coefficient in inverse meters.
    :param c_mts: The temperature and salinity-corrected attenuation coefficient in inverse meters.
    """
    daq_time: datetime
    a_wavelength: tuple[float, ...]
    c_wavelength: tuple[float, ...]
    temperature: float
    salinity: float
    a_mts: ArrayLike
    c_mts: ArrayLike

    def to_dict(self) -> dict:
        """
        Export the TSCorrectedPacket as a dictionary.

        :return: A dict representation of the TSCorrectedPacket, where keys are attributes and values are attr values.
        """
        return self._asdict()

    def to_xarray(self) -> xr.Dataset:
        """
        Export the TSCorrectedPacket as an xarray.Dataset.

        :return: An xarray.Dataset representation of the TSCorrectedPacket.
        """
        ds = xr.Dataset()
        ds = ds.assign_coords({'daq_time': self.daq_time,
                               'a_wavelength': np.array(self.a_wavelength),
                               'c_wavelength': np.array(self.c_wavelength)})
        ds['temperature'] = self.temperature
        ds['salinity'] = self.salinity
        ds['a_mts'] = (['daq_time', 'a_wavelength'], [self.a_mts])
        ds['c_mts'] = (['daq_time', 'c_wavelength'], [self.c_mts])
        return ds
